
from os import chmod
from basics import *
from numpy.core.numerictypes import *
import numpy
import sys
from FITS import *
from datetime import datetime,timedelta
import ephem
from time import time

class DPErecord:
    def __init__(self,record,lampkey,flatkey,fringekey,slitkey,skykey):
        self.record = record
        self.filename,self.inst,self.date,self.start,self.mjd,self.etime,hilo = self.record.split()[:7]
        self.object = join(self.record.split()[7:],"").lower()
        self.object = self.object.replace("(","_")
        self.object = self.object.replace(")","_")
        self.object = self.object.replace("[","_")
        self.object = self.object.replace("]","_")
        self.object = self.object.replace("#","_")
        self.object = self.object.replace("/","_")
        self.object = self.object.replace(":","_")
        self.dirname = os.path.dirname(self.filename)
        self.basename = os.path.basename(self.filename)
        self.hilo = (hilo == "HI")

        self.mjd = float(self.mjd)
#        month,day,year = ephem.toGregorian(self.mjd)
#        day = int(day+1.0)
#        self.date = "%4d-%2s-%2s" % (year,zfill(month,2),zfill(day,2))
        self.etime = float(self.etime)
#        self.mjd += 0.5*(self.etime/3600.0)/24.0
        self.last = None
        self.newname = None
        self.simple = None
        self.type = []

        if skykey and skykey in self.object:
           self.type.append("sky"); return

        if lampkey in self.object:
           self.type.append("lamp"); return

        if flatkey in self.object:
           self.type.append("flat"); return

        if slitkey in self.object:
           self.type.append("slit")

        if fringekey and fringekey in self.object and flatkey not in self.object:
           self.type.append("fringe")

        if not len(self.type): self.type.append("other")

        if self.type == ["slit"]:
           if fringekey and fringekey in self.object: self.type.append("fringe")

        if self.etime == 0.0: self.type = ["zero"]
    def __repr__(self):
         return "%s(type=%r)\n" % (self.object,self.type)

class DPEdb:
    def __init__(self,dbfile,udate,ignore,lampkey,flatkey,fringekey,slitkey,skykey):
        self.ignore = ignore
        self.flatkey = flatkey
        self.slitkey = slitkey
        self.fringekey = fringekey
        self.lampkey = lampkey
        self.skykey = skykey
        self.file = open(dbfile,"r")
        self.read(udate,lampkey,flatkey,fringekey,slitkey,skykey)
        self.file.close()
    def read(self,udate,lampkey,flatkey,fringekey,slitkey,skykey):
        self.records = [DPErecord(rec,lampkey,flatkey,fringekey,slitkey,skykey) for rec in self.file.xreadlines() if rec[0] != "#" and len(rec.split()) > 7]
        nrec = len(self.records)
        if udate:
           use_date = [((self.records[j].date == udate) and ("other" in self.records[j].type)) for j in range(nrec)]
        else:
           use_date = ones((nrec,),float32)
        use_calib = [("other" not in self.records[j].type) for j in range(nrec)]
        self.records = [self.records[j] for j in range(nrec) if (use_date[j] or use_calib[j]) and self.ignore not in self.records[j].object]

CALIBTYPES = ["lamp","flat","slit","fringe"]

class DPEobservations:
    def __init__(self,target,owner,runid,udate,dbdata,onlyflat,onlyslit,onlyfringe,onlylamp,twodim,logb,clean,noslitfn,forked,dkx=1.0,noobjects=0,fake=0,rn=0,combine=1,skyshift=1,orderspan=None,noskyaps=None,noskysub=None,masksky=None):
        self.pipeline = None
        self.owner = owner
        self.runid = runid
        self.newlamp = False
        self.newflat = False
        self.newslit = False
        self.newfringe = False
        self.target = target.lower()
        self.prereq = "targ_lamp"
        self.udate = udate
        self.dbdata = dbdata
        self.onlyflat = onlyflat
        self.onlyslit = onlyslit
        self.onlylamp = onlylamp
        self.onlyfringe = onlyfringe
        self.twodim = twodim
        self.skyshift = skyshift
        self.fake = fake
        self.noslitfn = noslitfn
        self.clean = clean
        self.combine = combine
        self.forked = forked
        self.dkx = dkx
        self.orderspan = orderspan
        self.noobjects = noobjects
        self.noskyaps = noskyaps
        self.noskysub = noskysub
        self.masksky = masksky
        self.logb = logb
        self.rn = rn
        self.IdentifyObject()
        if self.onlyfringe: self.target = "fringe"
        if self.onlyflat: self.target = "flat"
        if self.onlylamp: self.target = "lamp"
        if self.onlyslit: self.target = "slit"
        self.usefringe = None
        if self.onlylamp: self.spectra = self.spectra[:1]
        if len(self.spectra):
            self.useflat = "flat.fits"
            self.uselamp = "lamp.fits"
            self.useslit = "slit.fits"
            self.useblaze = "blaze.fits"
            self.useslits = [s for s in glob("slit??????.fits")]
            self.useflats = [s for s in glob("flat??????.fits")]

            if self.udate:
               self.dir = "%s-%s" % (self.target,self.udate)
               self.prefix = "%s-%s" % (self.target,self.udate)
               self.pipeline = "%s-%s.pipeline" % (self.target,self.udate)
            else:
               self.dir = "%s" % (self.target)
               self.prefix = "%s" % (self.target)
               self.pipeline = "%s.pipeline" % (self.target)

            if not os.path.exists(self.dir): os.mkdir(self.dir)

            if not os.path.exists("./Final-Products") and not os.path.islink("./Final-Products"):
              os.mkdir("./Final-Products")

            if not os.path.exists(self.useflat) and not self.onlyflat and not os.path.islink(self.useflat) and not self.onlyslit and not self.onlyfringe:
               self.AssociateCalib( self.spectra[0],"flat")
               self.newflat = True
               self.flatsetup = DPEobservations(self.dbdata.flatkey,self.owner,self.runid,None,self.dbdata,1,0,0,0,0,0,0,0,0,orderspan=self.orderspan)
               self.flatsetup.prereq = ""

            if not self.onlyflat and self.dbdata.fringekey:
               self.usefringe = "fringe.fits"
               if not os.path.exists(self.usefringe) and not self.onlyfringe and not os.path.islink(self.usefringe) and not self.onlyslit:
                  self.AssociateCalib( self.spectra[0],"fringe")
                  self.newfringe = True
                  self.fringesetup = DPEobservations(self.dbdata.fringekey,self.owner,self.runid,None,self.dbdata,0,0,1,0,0,0,0,0,0,orderspan=self.orderspan)

            if not self.onlyflat and not os.path.exists(self.useslit) and not self.onlyslit and not os.path.islink(self.useslit) and not self.onlyfringe:
               self.AssociateCalib( self.spectra[0],"slit")
               self.newslit = True
               self.slitsetup = DPEobservations(self.dbdata.slitkey,self.owner,self.runid,None,self.dbdata,0,1,0,0,0,0,0,self.noslitfn,0,orderspan=self.orderspan)
               self.slitsetup.prereq = "targ_flat"

            if not self.onlyflat and not self.onlyslit and not self.onlyfringe:
               if not os.path.exists(self.useflat) and not os.path.islink(self.useflat):
                  print "problem with flatfield?";exit()

               if not os.path.exists(self.uselamp) and not self.onlylamp and not os.path.islink(self.uselamp) and not self.onlyslit:
                  self.AssociateCalib( self.spectra[0],"lamp")
                  self.newlamp = True
                  self.lampsetup = DPEobservations(self.dbdata.lampkey,self.owner,self.runid,None,self.dbdata,0,0,0,1,0,0,0,1,0,orderspan=self.orderspan)
                  self.lampsetup.prereq = "targ_slit"

                  if not self.onlyfringe and not self.onlylamp:
                     if not os.path.exists(self.uselamp) and not os.path.islink(self.uselamp):
                        print "problem with lamp master frame?";exit()

                  if not self.onlyfringe and self.dbdata.fringekey:
                     if not os.path.exists(self.usefringe) and not os.path.islink(self.usefringe):
                        print "problem with fringe frame?";exit()

            os.chdir(self.dir)
            if not self.onlyflat:
               if not os.path.exists(self.useflat) and not os.path.islink(self.useflat):
                  os.symlink(join(["../",self.useflat],""),self.useflat)
               if not os.path.exists(self.useblaze) and not os.path.islink(self.useblaze):
                  os.symlink(join(["../",self.useblaze],""),self.useblaze)
               if not self.onlyfringe and self.dbdata.fringekey:
                 if not os.path.exists(self.usefringe) and not os.path.islink(self.usefringe):
                    os.symlink(join(["../",self.usefringe],""),self.usefringe)
               if not os.path.exists(self.uselamp) and not os.path.islink(self.uselamp) and not self.onlylamp:
                  os.symlink(join(["../",self.uselamp],""),self.uselamp)
               [self.AssociateCalib(n,"lamp") for n in self.spectra]

            [self.AssociateCalib(n,"slit") for n in self.spectra]
            [self.AssociateCalib(n,"flat") for n in self.spectra]

            if self.dbdata.skykey:
               [self.AssociateCalib(n,"sky") for n in self.spectra]

            self.OpenPipeline()
            self.AccumulateSteps()
            self.DumpPipeline()
            os.chdir("../")

    def IdentifyObject(self):
        self.spectra = nonzero([(self.target in rec.object and \
                (sum([r not in CALIBTYPES for r in rec.type]) or \
                (self.onlyflat and "flat" in rec.type) or \
                (self.onlylamp and "lamp" in rec.type) or \
                (self.onlyslit and "slit" in rec.type) or \
                (self.onlyfringe and "fringe" in rec.type))) for rec in self.dbdata.records])

    def AssociateCalib(self,n,ctype):
        nrec = arange(len(self.dbdata.records))
        self.dbdata.records[n].usesky = -99
        record = self.dbdata.records[n].record
        fname = self.dbdata.records[n].filename
        ftype = self.dbdata.records[n].type
        object = self.dbdata.records[n].object
        mjd = self.dbdata.records[n].mjd
        etime  = self.dbdata.records[n].etime
        date = self.dbdata.records[n].date
        inst = self.dbdata.records[n].inst
#        slita = self.dbdata.records[n].slita
        use = [(rec.inst == inst and ctype in rec.type) for rec in self.dbdata.records]
#        dmjd = [abs(rec.mjd - mjd) + (abs(rec.slita-slita)/3600)  + 1e38 * (rec.inst != inst) + 1e38*(ctype not in rec.type) for rec in self.dbdata.records]
        dmjd = [abs(rec.mjd - mjd) + 1e38 * (rec.inst != inst) + 1e38*(ctype not in rec.type) for rec in self.dbdata.records]
        dmjd = where(equal(nrec,n),1e38,dmjd)
        if ctype == "sky" and "sky" not in ftype:
           usesky = [j for j,rec in enumerate(self.dbdata.records) \
                if object in rec.object and "sky" in rec.type]
           self.dbdata.records[n].usesky = -99
           if len(usesky):
             self.dbdata.records[n].usesky = usesky[0]
           self.dbdata.records[n].skydmjd = min(dmjd)
        if ctype == "lamp":
           if self.onlylamp:
              self.dbdata.records[n].uselamp = n
           else:
              self.dbdata.records[n].uselamp = argmin(dmjd)
              self.dbdata.records[n].lampdmjd = min(dmjd)
        elif ctype == "flat":
           self.dbdata.records[n].useflat = nonzero(use)
        elif ctype == "fringe":
           self.dbdata.records[n].usefringe = nonzero(use)
        elif ctype == "slit":
           if self.onlyslit:
              self.dbdata.records[n].useslit = n
           else:
              self.dbdata.records[n].useslit = argmin(dmjd)
              self.dbdata.records[n].slitdmjd = min(dmjd)

    def OpenPipeline(self):
        self.pipelines = []
        self.stages = []
        self.newfiles = []

    def DumpPipeline(self):
        print "Opening",self.pipeline
        self.pipe = open( self.pipeline,"w")
        self.pipe.write("\n# Pipeline for Target %s\n\n" % (self.target))
        if self.onlyflat or self.onlyfringe: self.stages.append("clean")
#        if not self.onlyflat: self.pipe.write("include ../GAIN.dat\n\n")
        if not self.onlyflat:
           self.pipe.write("GAIN := $(shell python -c \'import FITS;f=FITS.FITS(\"flat.fits\");print f[\"newgain\"];f.close()\')\n")
#        self.pipe.write("all: %s\n\n" % (join(self.stages)))
        self.pipe.write("all: %s\n\n" % (join(self.stages+["../targ_%s" % (self.target)])))
        self.pipe.writelines(self.pipelines)
        self.pipe.write("../targ_%s:\n" % (self.target))
        self.pipe.write("\ttouch ../targ_%s\n\n" % (self.target))
        self.pipe.write("clean:\n")
        cleanup = []
        [[cleanup.append(f) for f in newfs if f not in cleanup if self.target in f] for newfs in self.newfiles]
        if len(cleanup): self.pipe.write("\trm -f %s\n\n" % (join(cleanup)))
        self.pipe.write("veryclean:\n")
        verycleanup = []
        [[verycleanup.append(f) for f in newfs if f not in verycleanup if self.target in f] for newfs in self.newfiles]
        if not self.onlyflat and not self.onlyfringe: [verycleanup.append(stage) for stage in self.stages]
        if len(cleanup): self.pipe.write("\trm -f %s\n\n" % (join(verycleanup)))
        try: chmod(self.pipe.name,493)
        except: pass
        self.pipe.close()
        if not os.path.exists("Makefile"): os.symlink(self.pipeline,"Makefile")

    def InitStage(self,stage,deps=[],sub=0):
        if sub: oldstage = self.stages[-2]
        elif len(self.stages): oldstage = self.stages[-1]
        else: oldstage = ""
        self.stage = "stage-%s" % (stage)
        self.pipelines.append( "%s:" % (self.stage))
        if len(self.stages): self.pipelines[-1] += " %s" % (oldstage)
        if deps: self.pipelines[-1] += " stage-"+" stage-".join(deps)
        self.pipelines[-1] += "\n"
        if not sub: self.stages.append(self.stage)

    def EndStage(self,extras=[]):
        stage = self.stage
        self.pipelines.append( "\ttouch %s\n" % (stage))
        self.pipelines.append( "\n")
        newfiles = extras
        for n in self.spectra:
           newfiles.append( self.dbdata.records[n].newname)
           if "uselamp" in dir(self.dbdata.records[n]):
              m = self.dbdata.records[n].uselamp
              if m >= 0 and self.dbdata.records[m].newname not in newfiles:
                newfiles.append( self.dbdata.records[m].newname)
           if "useslit" in dir(self.dbdata.records[n]):
              m = self.dbdata.records[n].useslit
              if m >= 0 and self.dbdata.records[m].newname not in newfiles:
                newfiles.append( self.dbdata.records[m].newname)
        self.newfiles.append([f for f in newfiles if type(f) == StringType])

    def WriteLine(self,line="",stage=None):
        if line:
            if line[0] == "#":
               self.pipelines.append("%s\n" % (line))
               self.pipelines.append("\techo Stage: %s\n" % (line[1:]))
            else:
               self.pipelines.append("\t%s\n" % (line))
        else: self.pipelines.append("\n")

    def AccumulateSteps(self):
        if self.onlyflat:
           self.InitStage("overscan")
           self.WriteLine("# Overscan Subtraction, Trimming, and Image Reorientation.")
           self.Overscan( self.dbdata.records[self.spectra[0]].useslit,"slit")
           for n in self.spectra:
              self.Overscan(n,"flat")
           self.EndStage()

           if len(self.spectra) > 1:
               self.InitStage("gain")
               self.WriteLine("# Compute the gain.")
               self.ComputeGain(self.spectra)
               self.EndStage()

           self.InitStage("ydist")
           self.WriteLine("# Measuring the Y Distortion.")
           for n in self.spectra:
              self.YDistortion( self.dbdata.records[self.spectra[0]].useslit)
           self.EndStage()

           self.InitStage("ydist-copy")
           self.WriteLine("# Copying the Y Distortion.")
           for n in self.spectra:
              self.Copyrect("-ydist",n,self.dbdata.records[self.spectra[0]].useslit)
           self.EndStage()

           self.InitStage("normalize")
           self.WriteLine("# Normalizing the Flatfields.")
           [self.NormFlat(n) for n in self.spectra]
           self.EndStage( [self.dbdata.records[n].newname.replace("flt.fits","blz.fits") for n in self.spectra])

           self.InitStage("combine")
           self.WriteLine("# Combine the Flats.")
           flists = self.GroupFlats(self.spectra)
#           self.CombineFlats(self.spectra,"_flat2d")
           self.useflats = []
           self.useblazes = []
           [self.CombineFlats(flist,"_flat2d") for flist in flists]
           self.EndStage()

           if len(self.useflats) > 1:
              for useflat in self.useflats:
                  if not os.path.exists( join(["../",useflat],"")) and not os.path.islink( join(["../",useflat],"")):
                      os.symlink( os.getcwd()+"/"+useflat, join(["../",useflat],""))
           if not os.path.exists( join(["../",self.useflat],"")) and not os.path.islink( join(["../",self.useflat],"")):
               os.symlink( os.getcwd()+"/"+self.useflats[0], join(["../",self.useflat],""))
           if len(self.useblazes) > 1:
              for useblaze in self.useblazes:
                  if not os.path.exists( join(["../",useblaze],"")) and not os.path.islink( join(["../",useblaze],"")):
                      os.symlink( os.getcwd()+"/"+useblaze, join(["../",useblaze],""))
           if not os.path.exists( join(["../",self.useblaze],"")) and not os.path.islink( join(["../",self.useblaze],"")):
               os.symlink( os.getcwd()+"/"+self.useblazes[0], join(["../",self.useblaze],""))

           return

        if self.onlyfringe:
           self.InitStage("overscan")
           self.WriteLine("# Overscan Subtraction, Trimming, and Image Reorientation.")
           for n in self.spectra:
              self.Overscan(n,"fringe")
           self.EndStage()

           self.InitStage("flatten")
           self.WriteLine("# Flattening.")
           [self.Flatten(n) for n in self.spectra]
           self.EndStage()

           self.InitStage("ydist-copy")
           self.WriteLine("# Copying the Y Distortion.")
           self.Maprect()
           for n in self.spectra:
              self.Copyrect("-ydist",n)
           self.EndStage()

           self.InitStage("orders-copy")
           self.WriteLine("# Copy the Order Edges.")
           for n in self.spectra:
              if n == self.spectra[0]: self.CopyOrders(n,arc=1)
              else: self.CopyOrders(n,self.spectra[0],arc=0)
           self.EndStage()

           self.InitStage("wdist-copy")
           for n in self.spectra:
              self.Copyrect("-wdist",n,None)
           self.WriteLine()

           if not self.noslitfn:
               self.InitStage("deslitfn")
               self.WriteLine("# Dividing out the 2D Image of the Slit Function.")
               for n in self.spectra:
                  self.DeSlit(n,n)
               self.EndStage()

           self.InitStage("getfringes")
           self.WriteLine("# Get Fringes.")
           [self.GetFringes(n) for n in self.spectra]
           self.EndStage()

           self.InitStage("combine")
           self.WriteLine("# Combine the Fringe Frames.")
           self.CombineFlats(self.spectra,"z")
           self.EndStage()

           if not os.path.exists( join(["../",self.usefringe],"")) and not os.path.islink( join(["../",self.usefringe],"")):
               os.symlink( os.getcwd()+"/"+self.usefringe, join(["../",self.usefringe],""))

           return

        if self.onlyslit:

           self.InitStage("overscan")
           self.WriteLine("# Overscan Subtraction, Trimming, and Image Reorientation.")
           for n in self.spectra:
              self.Overscan(n,"slit")
           self.EndStage()

           self.InitStage("combine")
           self.WriteLine("# Combine the Slit Frames.")
#           diff = lambda l: [abs(l[i]-l[i+1]) for i in range(len(l)-1)]
##           [sys.stdout.write("%d "%(i)) for i in range(len(diff(self.spectra))) if diff(self.spectra)[i] > 1]
#           dends = [(i+1) for i in range(len(diff(self.spectra))) if diff(self.spectra)[i] > 1]
#           slists = []
#           if len(dends):
#              for i in range(len(dends)):
#                  if i == 0:
#                     slists.append(self.spectra[:dends[i]])
#                     slists.append(self.spectra[dends[i]:dends[i+1]])
#                  elif i == range(len(dends))[-1]:
#                     slists.append(self.spectra[dends[i]:])
#                  else:
#                     slists.append(self.spectra[dends[i]:dends[i+1]])
#           else: slists.append(self.spectra)
           slists = self.GroupFlats(self.spectra)
           [self.CombineFlats(slist,"") for slist in slists]
           self.EndStage()
#           self.spectra = self.spectra[:1]
           self.spectra = [slist[0] for slist in slists]

           self.InitStage("flatten")
           self.WriteLine("# Flattening.")
           self.MakeMask()
           for n in self.spectra:
              self.Flatten(n)
           self.EndStage()

           self.InitStage("ydist")
           self.WriteLine("# Measuring the Y Distortion.")
           for n in self.spectra:
              self.YDistortion( n)
           self.EndStage()

           self.InitStage("orders")
           self.WriteLine("# Find the Order Edges.")
           for n in self.spectra:
              self.FindOrders( n)
           self.EndStage()

           slitnames = []
           if self.noslitfn:
#              slitname = self.dbdata.records[self.spectra[0]].newname
              [os.symlink(self.dbdata.records[spec].newname,self.dbdata.records[spec].newname.replace(".fits","_slitfn.fits")) for spec in self.spectra if not os.path.islink(self.dbdata.records[spec].newname.replace(".fits","_slitfn.fits"))]
              [slitnames.append(self.dbdata.records[spec].newname.replace(".fits","_slitfn.fits")) for spec in self.spectra]
           else:
              self.InitStage("slitfn")
              self.WriteLine("# Get the 2D Image of the Slit Function.")
              for n in self.spectra:
                 self.GetSlitFunction( n)
              self.EndStage()
#              slitname = self.dbdata.records[self.spectra[0]].newname.replace(".fits","_slitfn.fits")
              [slitnames.append(self.dbdata.records[spec].newname.replace(".fits","_slitfn.fits")) for spec in self.spectra]

           if len(slitnames) == 1:
              slitname = slitnames[0]
              if os.path.exists( self.useslit) or os.path.islink(self.useslit): os.remove(self.useslit)
              os.symlink(slitname,self.useslit)
             
              if not os.path.exists( join(["../",self.useslit],"")) and not os.path.islink( join(["../",self.useslit],"")):
                  os.symlink( os.getcwd()+"/"+self.useslit, join(["../",self.useslit],""))
           else:
              for slitname in slitnames:
                  slitlink = slitname.replace("median","slit").replace("ff_slitfn","")
                  if os.path.exists( slitlink) or os.path.islink(slitlink): os.remove(slitlink)
                  os.symlink(slitname,slitlink)
                  if os.path.exists( self.useslit) or os.path.islink(self.useslit): os.remove(self.useslit)
                  os.symlink(slitname,self.useslit)
             
                  if not os.path.exists( join(["../",slitlink],"")) and not os.path.islink( join(["../",slitlink],"")):
                      os.symlink( os.getcwd()+"/"+slitlink, join(["../",slitlink],""))
                  if not os.path.exists( join(["../",self.useslit],"")) and not os.path.islink( join(["../",self.useslit],"")):
                      os.symlink( os.getcwd()+"/"+self.useslit, join(["../",self.useslit],""))

           return

        self.InitStage("overscan")
        self.WriteLine("# Overscan Subtraction, Trimming, and Image Reorientation.")
        for n in self.spectra:
           if self.onlylamp:
              self.Overscan(n,"lamp")
           else:
              self.Overscan(n,"spec")
           self.Overscan( self.dbdata.records[n].uselamp,"lamp")
        self.EndStage()

        self.InitStage("flatten")
        self.WriteLine("# Flattening.")
        for n in self.spectra:
           self.Flatten(n)
           self.Flatten( self.dbdata.records[n].uselamp)
        self.EndStage()

        if self.dbdata.fringekey and not self.onlylamp:
            self.InitStage("defringe")
            self.WriteLine("# De-fringing.")
            for n in self.spectra:
               self.DeFringe(n)
               self.DeFringe( self.dbdata.records[n].uselamp)
            self.EndStage()

        self.InitStage("ydist-copy")
        self.WriteLine("# Copying the Y Distortion.")
#        self.Maprect(self.spectra[0])
        slitdates, mspecs = [],[]
#        [specdates.append(self.dbdata.records[n].date) for n in self.spectra if self.dbdata.records[n].date not in specdates]
        for n in self.spectra:
            sdate = self.AssociateFlat(n,ftype="slit")
            if sdate not in slitdates:
               slitdates.append(sdate)
               mspecs.append(n)
        [self.Maprect(n) for n in mspecs]
        for n in self.spectra:
           self.Copyrect("-ydist",n)
           self.Copyrect("-ydist",self.dbdata.records[n].uselamp)
        self.EndStage()

        self.InitStage("orders-copy")
        self.WriteLine("# Copy the Order Edges.")
        for n in self.spectra:
           self.CopyOrders(self.dbdata.records[n].uselamp,arc=1)
           self.CopyOrders(n,self.dbdata.records[n].uselamp)
        self.EndStage()

        if not self.noslitfn:
            self.InitStage("deslitfn")
            self.WriteLine("# Dividing out the 2D Image of the Slit Function.")
            for n in self.spectra:
               self.DeSlit(self.dbdata.records[n].uselamp,n)
            for n in self.spectra:
               self.DeSlit(self.dbdata.records[n].uselamp,self.dbdata.records[n].uselamp)
            self.EndStage()

        self.InitStage("xdist")
        self.WriteLine("# Get X Distortion.")
        for n in self.spectra:
           self.XDistortion(self.dbdata.records[n].uselamp)
        self.EndStage()

        self.InitStage("xdist-copy")
        self.WriteLine("# Copying X Distortion.")
        for n in self.spectra:
           self.Copyrect("-xdist",n,self.dbdata.records[n].uselamp)
           self.Copyrect("-xdist",self.dbdata.records[n].uselamp,self.dbdata.records[n].uselamp)
        self.EndStage()

        self.InitStage("wdist")
        self.WriteLine("# Do the Wavelength Calibration.")
        for n in self.spectra:
           self.MatchOrders(self.dbdata.records[n].uselamp)
        self.WriteLine()

        for n in self.spectra:
           self.FindLines(self.dbdata.records[n].uselamp,th=10)
        self.WriteLine()

        for n in self.spectra:
           self.MatchLamps(self.dbdata.records[n].uselamp,master= not self.onlylamp)
        self.EndStage()

        self.InitStage("wdist-copy")
        for n in self.spectra:
           self.CopyECOrders(self.dbdata.records[n].uselamp,self.dbdata.records[n].uselamp)
        self.WriteLine()

        for n in self.spectra:
           self.CopyECOrders(n,self.dbdata.records[n].uselamp)
           self.CopyECOrders(self.dbdata.records[n].uselamp,self.dbdata.records[n].uselamp)
        self.WriteLine()

        for n in self.spectra:
           self.Copyrect("-wdist",n,self.dbdata.records[n].uselamp)
           self.Copyrect("-wdist",self.dbdata.records[n].uselamp,self.dbdata.records[n].uselamp)
        self.WriteLine()

        for n in self.spectra:
           self.Copyrect("-edist",n,self.dbdata.records[n].uselamp)
           self.Copyrect("-edist",self.dbdata.records[n].uselamp,self.dbdata.records[n].uselamp)
        self.EndStage()

        if self.onlylamp:
           if not os.path.exists( self.uselamp) and not os.path.islink( self.uselamp):
               os.symlink( self.dbdata.records[self.spectra[0]].newname, self.uselamp)
           if not os.path.exists( join(["../",self.uselamp],"")) and not os.path.islink( join(["../",self.uselamp],"")):
               os.symlink( os.getcwd()+"/"+self.uselamp, join(["../",self.uselamp],""))
           return

        if self.skyshift:
            self.InitStage("wdist-offset")
            self.WriteLine("# Generate 1D Spectra of the Sky")
            for n in self.spectra:
               self.Get1DSky(n)
            self.WriteLine()

            self.WriteLine("# Generate Database of Individual Sky Lines")
            for n in self.spectra:
               self.FindLines(n)
            self.WriteLine()

            self.WriteLine("# Identify and Fit Positions of Sky Lines")
            for n in self.spectra:
               self.MatchSky(n)
            self.EndStage()

        if not self.noskyaps and not self.noskysub:
           self.InitStage("skyaps")
           self.WriteLine("# Generate Sky Apertures.")
           self.SkyMask( self.spectra)
#           if self.noobjects: self.SkyMask( self.spectra)
#           else: self.SkyAps( self.spectra)
           self.EndStage()

        if [n for n in self.spectra if (self.dbdata.records[n].usesky >= 0 or "sky" in self.dbdata.records[n].type)]:
           self.InitStage("crclean")
           self.WriteLine("# Run pyrclean.")
           [self.CRClean(n) for n in self.spectra if (self.dbdata.records[n].usesky >= 0 or "sky" in self.dbdata.records[n].type)]
           self.EndStage()

        if not self.noskysub:
           self.InitStage("skyrect")
           self.WriteLine("# Determine the Sky Background.")
           for n in self.spectra:
             if self.dbdata.records[n].usesky >= 0: self.SkyExposure(n, self.dbdata.records[n].usesky)
             elif "other" in self.dbdata.records[n].type: self.SkyModel(n)
           self.EndStage()
          
           self.spectra = [n for n in self.spectra if "other" in self.dbdata.records[n].type]
          
           self.InitStage("skysub")
           self.WriteLine("# Subtract the Sky Background.")
           for n in self.spectra:
              self.SkySubtract(n)
           self.EndStage()
        else:
           self.InitStage("scatrect")
           self.WriteLine("# Determine the Scattered Light.")
           [self.ScatModel(n) for n in self.spectra]
           self.EndStage()
          
           self.spectra = [n for n in self.spectra if "other" in self.dbdata.records[n].type]
          
           self.InitStage("subscat")
           self.WriteLine("# Subtract the Sky Background.")
           for n in self.spectra:
              self.SkySubtract(n)
           self.EndStage()


        if self.masksky:
           self.InitStage("mkskymask")
           self.WriteLine("# Generate a mask of the sky.")
           mim = lambda n: self.dbdata.records[n].newname.replace(".fits","or_sm.fits")
           [self.MakeSkyMask(n) for n in self.spectra]
           self.EndStage()

           self.InitStage("skyrect-2")
           self.WriteLine("# Determine the Sky Background Again.")
           [self.SkyModel(n,mask=mim(n)) for n in self.spectra]
           self.EndStage()
            
           self.InitStage("skysub-2")
           self.WriteLine("# Subtract the New Sky Background.")
           [self.SkySubtract(n,p=2) for n in self.spectra]
           self.EndStage()

        self.InitStage("objaps")
        self.WriteLine("# Generate Object Positions.")
        if self.noobjects:
           [self.PosObjects(self.spectra[n:n+1]) for n in range(len(self.spectra))]
        else:
           self.PosObjects(self.spectra)
           self.AllPos(self.spectra)
        self.EndStage()

        if self.twodim:
            if self.noobjects:
                self.InitStage("multiunrect")
                self.WriteLine("# Run Multiunrect.")
                self.Unrect(self.spectra)
#                self.Unrect(self.spectra[:1])
#                [self.Unrect([n],template=self.spectra[0]) for n in self.spectra[1:]]
                self.CollateRect()
                self.EndStage()

                self.InitStage("multispec")
                self.WriteLine("# Run extractmultispec.")
                self.MultiExtract(100.0)
                self.EndStage()

                self.InitStage("update-header")
                self.WriteLine("# Update header of extraction file.")
                self.AddHeaderInfo()
                self.EndStage()

                self.InitStage("finalize")
                self.WriteLine("# Copy Final Products.")
                self.CopyProducts()
                self.EndStage()

            else:
                self.InitStage("multiunrect")
                self.WriteLine("# Run Multiunrect.")
                self.Unrect(self.spectra)
                self.CollateRect()
                self.EndStage()

                self.InitStage("multispec")
                self.WriteLine("# Run extractmultispec.")
                if not self.noobjects: self.MultiExtract(100.0)
                self.EndStage()

                self.InitStage("update-header")
                self.WriteLine("# Update header of extraction file.")
                self.AddHeaderInfo()
                self.EndStage()

                self.InitStage("finalize")
                self.WriteLine("# Copy Final Products.")
                self.CopyProducts()
                self.EndStage()

        else:
            self.InitStage("mkmask")
            self.WriteLine("# Make a quick mask image of bad pixels")
            self.MakeMask()
            self.EndStage()

            if self.combine:
               self.InitStage("extractall")
               self.WriteLine("# Gauss-Hermite-Legendre-BSpline Extractions")
               self.lasts = ()
               self.GHLBExtract()
               self.EndStage()
            else:
               self.InitStage("extractall", ["extractall%d" % (j) for j in range(len(self.spectra))])
               self.EndStage()
               self.lasts = ()
               for j,n in enumerate(self.spectra):
                 self.InitStage("extractall%d" % (j),sub=1)
                 self.GHLBExtract(n,self.spectra[0])
                 self.EndStage()

            self.InitStage("update-header")
            self.WriteLine("# Update header of extraction file.")
            self.AddHeaderInfo()
            self.EndStage()

            self.InitStage("finalize")
            self.WriteLine("# Copy Final Products.")
            self.CopyProducts()
            self.EndStage()

            self.InitStage("reporting")
            self.WriteLine("# Generate html and pdf reports.")
            self.Report()
            self.EndStage()

        return

    def Overscan(self,n,ctype):
        if self.dbdata.records[n].last != self.Overscan:
           image = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
           suffix = ctype
           u = os.path.basename("../"+self.dbdata.records[n].filename).replace(".gz","").replace(".fits","")
           for l in letters: u = u.replace(l,"")
           F = FITS(image,readonly=1)
           self.binning = int(F["binning"].split("x")[1])
           self.sbinning = int(F["binning"].split("x")[0])
           F.close()
           if ctype == "slit": linebias = 0
           else: linebias = 1
           if self.udate:
              new = "%s-%s_%s%s.fits" % (self.target,self.udate,suffix,u)
           else:
              new = "%s_%s%s.fits" % (self.target,suffix,u)
           command = "DPEbias","-c 3",image,new
           if "flat" in self.dbdata.records[n].type: command += "-nofudge",
           if linebias: command += "-linebias",
           if self.orderspan: command += "-orderspan %d,%d" % tuple(self.orderspan),
           if ctype == "slit": linebias = 0
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].last = self.Overscan
           self.dbdata.records[n].newname = new
           self.dbdata.records[n].simple = new

    def ComputeGain(self,nlist):
         enames = [self.dbdata.records[n].newname for n in nlist]
         etimes = [self.dbdata.records[n].etime for n in nlist]
         netimes = [etimes.count(et) for et in etimes]
         eqt = greater_equal(netimes,2)
         if sum(eqt):
            nutimes = nonzero(eqt)[0]
            use = nonzero(equal(etimes,etimes[nutimes]))
            m,n = use[:2]
            im1 = enames[m]
            im2 = enames[n]
            command = "getGain",im1,im2,"> GAIN.dat"
         else:
            gain = 1.05
            command = "echo %f > GAIN.dat" % (gain)
         command = join(command)
         self.WriteLine(command)

    def YDistortion(self,n,ns=4,ox=2,oy=2):
        if self.dbdata.records[n].last != self.YDistortion:
            ### values used for MIKE-blue
#            ox,oy = 1,3
            pw = 10
            niter = 1
#            nx,dy = 4,32
            nx,dy = 8,32
            nbg = 0
            image = self.dbdata.records[n].newname
            gimage = self.dbdata.records[n].newname.replace(".fits","gmag.fits")
#            command = "efits",image,"\"VTKGradMag(i1,1)\"",gimage
#            command = join(command)
#            self.WriteLine(command)
            command = "getrect","-ydist", image,"-normwt 0",
            command +=  "-niter %d" % (niter),
            command += "-nx %d" % (nx), "-dy %d" % (dy), "-x %d" % (ox),"-dump"
            command += "-nbg %d" % (nbg),"-nsub2 %d" % (ns), "-y %d" % (oy), "-peakwin %d" % (pw)
            command = join(command)
            self.WriteLine(command)
#            command = "copyrect","-ydist",gimage,image
#            command = join(command)
#            self.WriteLine(command)
            self.dbdata.records[n].last = self.YDistortion

    def Maprect(self,n):
        foundslit = self.AssociateFlat(n,ftype="slit")
#        command = join(["cp","../"+foundslit,foundslit])
#        self.WriteLine(command)
        command = join(["maprect","-ydist",foundslit])
        self.WriteLine(command)

    def Copyrect(self,d,n,m=None):
        if self.dbdata.records[n].last != (self.Copyrect,d):
           if d == "-ydist":
              if m != None: dist = self.dbdata.records[m].newname
#              else: dist = "slit.fits"
              else: dist = self.AssociateFlat(n,ftype="slit")
           elif d == "-wdist" and m == None:
              dist = "lamp.fits"
           else:
              dist = self.dbdata.records[m].usename
           image = self.dbdata.records[n].newname
           command = join(["copyrect",d,dist,image])
           self.WriteLine(command)
           if d in ["-wdist","-edist"] and "lamp" in image:
               bimage = self.dbdata.records[n].newname.replace(".fits","b.fits")
               command = join(["copyrect",d,dist,image])
               self.WriteLine(command)
           self.dbdata.records[n].last = (self.Copyrect,d)

    def FindOrders(self,n,minlen=11,pmin=0.2,pmax=999.0,oo=3):
        if self.dbdata.records[n].last != self.FindOrders:
           image = self.dbdata.records[n].newname
           oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
           F = FITS(oimage,readonly=1)
           if self.rn: rn = self.rn
           else: rn = F["enoise"]
           g = F["egain"]
           binning = int(F["binning"].split("x")[0])
           F.close()
           oo = 6
           cut = 990
           top,bot = 1490,0
           command = "findorders",image,"-d",image.replace(".fits","orders.ps/vcps"),"-rn %f" % (rn),"-g ${GAIN}"
#           command += "-len %d" % (minlen*2.0/binning),
           command += "-len %d" % (minlen/binning),
           command += "-min %f" % (pmin), #"-max %f" % (pmax),
           command += "-extend", "-oo %d" % (oo),
           command += "-cut %d" % (cut),
           command += "-top %d" % (top), "-bot %d" % (bot),
           command += "-usemed",
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].last = self.FindOrders

    def GetSlitFunction(self,n):
        if self.dbdata.records[n].last != self.GetSlitFunction:
           image = self.dbdata.records[n].newname
           maskimage = "mask.fits"
           command = "GetSlitFunction",image,"-mask",maskimage,"-ms 1.2"
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].last = self.GetSlitFunction

    def AssociateFlat(self,n,ftype):
#        origim = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
#        F = FITS(origim)
#        sdate = F["ut-date"]
#        stime = F["ut-time"]
#        F.close()
        fdate = self.dbdata.records[n].date
        ftime = self.dbdata.records[n].start
        fdt = datetime.strptime(" ".join([fdate,ftime]), '%Y-%m-%d %H:%M:%S')
        midnt = datetime.strptime(" ".join([fdate,'00:00:00']), '%Y-%m-%d %H:%M:%S')
        noon = datetime.strptime(" ".join([fdate,'12:00:00']), '%Y-%m-%d %H:%M:%S')
        oneday = timedelta(days=1)
        if between(fdt,midnt,noon): fdate = str(fdt - oneday).split()[0]
        fname = "%s%s.fits" % (ftype,fdate.replace("-","")[2:])
        if ftype == "flat": ffiles = self.useflats
        elif ftype == "slit": ffiles = self.useslits
        else:
           print "Invalid choice for 'ftype'. Must be either 'flat' or 'slit'"
           sys.exit()
        if len(ffiles):
           if fname in ffiles:
              foundflat = fname
           else:
              odate = int(fname.replace("%s"%(ftype),"").replace(".fits",""))
              fdates = asarray([int(f.replace("%s"%(ftype),"").replace(".fits","")) for f in ffiles])
              foundflat = "%s%d.fits" % (ftype,fdates[argmin(abs(fdates-odate))])
        else:
           foundflat = "%s.fits" % (ftype)
        if not os.path.islink(foundflat): os.symlink("../"+foundflat,foundflat)
        return foundflat

    def CopyOrders(self,n,m=None,arc=0):
        if self.dbdata.records[n].last != self.CopyOrders:
           image = self.dbdata.records[n].newname
           if m !=None: foundslit = self.dbdata.records[m].newname
#           else: foundslit = "slit.fits"
           else: foundslit = self.AssociateFlat(n,ftype="slit")
           if arc:
#              fw = 10*2.0/self.sbinning
              fw = 5/self.sbinning
              command = "copyslit","-fwhm %f" % (fw), "-ydist","-fft",foundslit,image
              command = join(command)
           else:
              command = "copyslit",foundslit,image
              command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].last = self.CopyOrders

    def XDistortion(self,n,nx=1,dy=3,nsub2=1,ox=0,oy=1,b=2,nwt=100,niter=1):
        if self.dbdata.records[n].last != self.XDistortion:

           oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
           F = FITS(oimage,readonly=1)
           if self.rn: rn = self.rn
           else: rn = F["enoise"]
           g = F["egain"]
           sbinning = int(F["binning"].split("x")[0])
           binning = int(F["binning"].split("x")[1])
           F.close()

           image = self.dbdata.records[n].newname
           imageb = image.replace(".fits","b.fits")
           command = "efits",image
           command += "\'i1-VTKMedian(i1,1,%d,numret=1)\'" % (50/sbinning+1),imageb
           command = join(command)
           self.WriteLine(command)
           command = "efits",imageb
           command += "\'i1-VTKMedian(i1,%d,%d,numret=1)\'" % (10/binning+1,2/sbinning+1),imageb
           command = join(command)
           self.WriteLine(command)
           command = "efits",imageb
           command += "\'1.49*median(abs(i1))\'",imageb.replace(".fits","_scat.fits")
           command = join(command)
           self.WriteLine(command)
           command = "efits -r",imageb,imageb.replace(".fits","_scat.fits")
           command += "\'bl,bs=bwt(i2);bw=less(i2,bl+100*bs)*ones(i1.shape,float32);smo=divz(VTKGauss(i1*bw,sigmax=1,sigmay=0,numret=1),VTKGauss(1.*bw,sigmax=1,sigmay=0,numret=1));result=where(bw,i1,smo)\'",imageb
           command = join(command)
           self.WriteLine(command)
           command = "efits",imageb
           command += "\'i1*between(i1,-5*1.49*median(transpose(abs(i1)))[::,NewAxis],65e3)\'",imageb
           command = join(command)
           self.WriteLine(command)

           self.dbdata.records[n].usename = imageb

           command = "getrect","-xdist","-ydist",imageb
           command += "-nx %d" % (nx), "-dy %d" % (dy/binning), "-x %d" % (ox),
           command += "-nsub2 %d" % (nsub2), "-y %d" % (oy)
           command += "-b %d" % (b/binning), "-normwt %f" % (nwt), "niter %d" % (niter)
           command = join(command)
           self.WriteLine(command)
           self.GlobalXDistortion(imageb)
           self.dbdata.records[n].last = self.XDistortion

    def GlobalXDistortion(self,image,maxo=2,s=11):
        command = "globalrect",image
        command += "-maxo %d" % (maxo),"-go","-s %f" % (s),"-d %s" % (image.replace(".fits","glob.ps/vcps"))
        command = join(command)
        self.WriteLine(command)
        self.MapXDistortion(image)

    def MapXDistortion(self,image,s=8):
        command = "maprect","-xdist",image
        command = join(command)
        self.WriteLine(command)

    def MatchSky(self,n):
        if self.dbdata.records[n].last != self.MatchSky:
           image1d = self.dbdata.records[n].usename
           image = self.dbdata.records[n].newname
           command = "DPEMatchSky",image1d,"-o",image,"-go"
           command = join(command)
           self.WriteLine(command)
           if self.clean and 0:
              command = "rm", image1d.replace(".fits","_EC???.fits")
              command = join(command)
              self.WriteLine(command)
              self.WriteLine()
           self.dbdata.records[n].last = self.MatchSky

    def MatchOrders(self,n):
        if self.dbdata.records[n].last != self.MatchOrders:
           image = self.dbdata.records[n].newname.replace(".fits","b.fits")
           command = "DPEMatchOrders",image
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].usename = image.replace(".fits","specs.fits")
           self.dbdata.records[n].last = self.MatchOrders

    def FindLines(self,n,fwhm=5.0,th=5):
        if self.dbdata.records[n].last != self.FindLines:
           image = self.dbdata.records[n].usename
           command = "DPEFindLines",image
           command += "-fwhm %f" % (fwhm/(self.binning*2.0)), "-th %f" % (th)
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].last = self.FindLines

    def MatchLamps(self,n,master=0):
        xo,oo=5,4
        if self.dbdata.records[n].last != self.MatchLamps:
           image = self.dbdata.records[n].usename
           command = "DPEMatchLamps",image
           if master: command += "-m lamp.fits",
           else: command += "-x %d" % (xo), "-o %d" % (oo)
           command = join(command)
           self.WriteLine(command)
           if self.clean and 0:
             command = "rm", image.replace(".fits","_EC???.fits")
             command = join(command)
             self.WriteLine(command)
           self.dbdata.records[n].last = self.MatchLamps

    def CopyECOrders(self,n,m):
        if self.dbdata.records[n].last != self.CopyECOrders:
           wimage = self.dbdata.records[m].usename
           image = self.dbdata.records[n].newname
           command = join(["copyecords",wimage,image])
           self.WriteLine(command)
           if "lamp" in image:
              bimage = self.dbdata.records[n].newname.replace(".fits","b.fits")
              command = join(["copyecords",wimage,bimage])
              self.WriteLine(command)
           self.dbdata.records[n].last = self.CopyECOrders

    def Get1DSky(self,n):
        if self.dbdata.records[n].last != self.Get1DSky:
           image = self.dbdata.records[n].newname
           command = "mikeGet1DOrders",image
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].usename = self.dbdata.records[n].newname.replace(".fits","specs.fits")
           self.dbdata.records[n].last = self.Get1DSky

    def MapYDistortion(self,s=8):
        image = self.ydistortion
        command = "maprect","-ydist",image
        command = join(command)
        self.WriteLine(command)

    def notNormFlat(self,n):
        image = self.dbdata.records[n].newname
        oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
        F = FITS(oimage,readonly=1)
        naxis1 = F["naxis1"]
        if self.rn: rn = self.rn
        else: rn = F["enoise"]
        g = F["egain"]
        F.close()
        s1 = (naxis1/16.0)
        s2 = 0.0
        command = "PullFlat -2D -s1 %f -s2 %f %s" % (s1,s2,image)
        self.WriteLine(command)
        self.dbdata.records[n].newname = image.replace(".fits","flt.fits")

    def NormFlat(self,n,cut=2.0):
        image = self.dbdata.records[n].newname
        oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
        F = FITS(oimage,readonly=1)
        if self.rn: rn = self.rn
        else: rn = F["enoise"]
        g = F["egain"]
        sbinning = int(F["binning"].split("x")[0])
        binning = int(F["binning"].split("x")[1])
        F.close()
        command = "procMilky",image,"-dx %d" % (71/(binning*2.0)),"-cut %f" % (cut)
        command = join(command)
        self.WriteLine(command)
        self.dbdata.records[n].newname = image.replace(".fits","flt.fits")

    def GetFringes(self,n):
        image = self.dbdata.records[n].newname
        oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
        F = FITS(oimage,readonly=1)
        if self.rn: rn = self.rn
        else: rn = F["enoise"]
        g = F["egain"]
        F.close()
        command = "GetFringes",image
        command = join(command)
        self.WriteLine(command)
        self.dbdata.records[n].newname = image.replace(".fits","z.fits")

    def GroupFlats(self,flats):
        diff = lambda l: [abs(l[i]-l[i+1]) for i in range(len(l)-1)]
#        [sys.stdout.write("%d "%(i)) for i in range(len(diff(self.spectra))) if diff(self.spectra)[i] > 1]
        dends = [(i+1) for i in range(len(diff(flats))) if diff(flats)[i] > 1]
        flists = []
        if len(dends):
           for i in range(len(dends)):
               if i == 0:
                  flists.append(flats[:dends[i]])
                  if len(dends) < 2:
                     flists.append(flats[dends[i]:])
                     break
                  else:
                     flists.append(flats[dends[i]:dends[i+1]])
               elif i == range(len(dends))[-1]:
                  flists.append(flats[dends[i]:])
               else:
                  flists.append(flats[dends[i]:dends[i+1]])
        else: flists.append(flats)
        return flists

    def CombineFlats(self,flats,suffix="flt"):
        n = flats[0]
        oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
        F = FITS(oimage,readonly=1)
        if self.rn: rn = self.rn
        else: rn = F["enoise"]
        g = F["egain"]
        fdate = F["ut-date"][2:].replace("-","")
        F.close()
        allimages = []
        simages = ["i%d" % (j+1) for j in range(len(flats))]
        limages = join(simages,",")
        limages1 = "i%d" % (len(flats)+1)
        limages2 = simages[-1]
        allimages = [self.dbdata.records[n].newname for n in flats]
        if "fl" in suffix:
            finalflat= "flat%s.fits" % (fdate)
            rmsflat= "flatrms%s.fits" % (fdate)
            command = ["efits","-p"]
            command += allimages
            command += ["\'median([%s])\'" % (limages), finalflat]
            command = join(command)
            self.WriteLine(command)
            command = ["addkeys","-f",finalflat,"newgain","`cat GAIN.dat`"]
            command = join(command)
            self.WriteLine(command)
            command = ["efits","-p"]
            command += allimages
            command += [finalflat,"\'1.49*median(abs(subtract([%s],%s)))\'" % (limages,limages1)]
            command += [rmsflat]
            command = join(command)
            self.WriteLine(command)
            finalblaze= "blaze%s.fits" % (fdate)
            command = ["efits","-p"]
            command += [self.dbdata.records[n].newname.replace("flt.fits","blz.fits") for n in flats]
            command += ["\'mean([%s])\'" % (limages), finalblaze]
            command = join(command)
            self.WriteLine(command)
#            self.useflat = finalflat
#            self.useblaze = finalblaze
            self.useflats.append(finalflat)
            self.useblazes.append(finalblaze)
        elif suffix == "":
            final= "median%s.fits" % (fdate)
            command = ["efits","-p"]
            command += allimages
            command += ["\'median([%s])\'" % (limages), final]
            command = join(command)
            self.WriteLine(command)
            self.dbdata.records[flats[0]].newname = final
        elif suffix == "z":
            finalflat= "fringe.fits"
            rmsflat= "fringerms.fits"
            command = ["efits","-p"]
            command += allimages
            command += ["\'divz(sum(allimages*greater(allimages,0)),sum(1.*greater(allimages,0)))\'", finalflat]
            command = join(command)
            self.WriteLine(command)
            self.useflat = finalflat
        self.WriteLine()

    def MakeMask(self):
        maskimage = "mask.fits"
        command = "efits",self.useflat,"\"1.0*greater(i1,0.25)\"",maskimage
        command = join(command)
        self.WriteLine(command)

    def Flatten(self,n):
        if self.dbdata.records[n].last != self.Flatten:
           image = self.dbdata.records[n].newname
           if len(self.useflats): foundflat = self.AssociateFlat(n,ftype="flat")
           else: foundflat = self.useflat
           command = "flat2d", foundflat, image,
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].newname = image.replace(".fits","f.fits")
           if self.onlyslit:
              image = self.dbdata.records[n].newname
              maskimage = "mask.fits"
              command = "flat2d", maskimage, image,
              command = join(command)
              self.WriteLine(command)
              self.dbdata.records[n].newname = image.replace(".fits","f.fits")
           self.dbdata.records[n].last = self.Flatten

    def DeFringe(self,n):
        if self.dbdata.records[n].last != self.DeFringe:
           image = self.dbdata.records[n].newname
           command = "flat2d", self.usefringe, image,
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].newname = image.replace(".fits","f.fits")
           self.dbdata.records[n].last = self.DeFringe

    def DeSlit(self,l,n,m=None):
        if self.dbdata.records[n].last != self.DeSlit:
           rimage = self.dbdata.records[l].newname
           image = self.dbdata.records[n].newname
           if m != None: slitfn  = self.dbdata.records[m].newname.replace(".fits","_slitfn.fits")
           else: slitfn  = "slit.fits"
           if rimage != image:
             command = "mikeflat1d", "-fft", slitfn, rimage, image,
           else:
             command = "mikeflat1d", "-fft", slitfn, image,
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].newname = image.replace(".fits","s.fits")
           self.dbdata.records[n].last = self.DeSlit

    def SkyAps(self,spectra,nsig=2.0,oo=5):
        if self.dbdata.records[spectra[0]].last != self.SkyAps:
           command = ["mkorderaps","-ydist","-nsig %f" % (nsig),"-oo %d" % (oo),]
#           if self.sbinning == 1: edge = 3
#           else: edge = 1
           edge = 1
           command += "-edge %d" % (edge),
           for n in spectra:
              command += self.dbdata.records[n].newname,
              self.dbdata.records[n].last = self.SkyAps
#           command += "-maskout",
           command = join(command)
           self.WriteLine(command)

    def SkyMask(self,spectra,nsig=2.5):
        if self.dbdata.records[spectra[0]].last != self.SkyMask:
           for n in spectra:
              command = "bsky -max 55e3 -nx 32 -ny 32 -simple %s" % ( self.dbdata.records[n].newname)
              self.WriteLine(command)
              command = ("efits", self.dbdata.records[n].newname,
                        self.dbdata.records[n].newname.replace(".fits","m.fits"),
                        self.dbdata.records[n].newname.replace(".fits","s.fits"),
                        "\"VTKLessEqual(VTKConvolve(VTKGreaterEqual(VTKDivide(VTKSubtract(i1,i2),i3),%f),1,3),0.25)\"" % (nsig),
                        self.dbdata.records[n].newname.replace(".fits","q.fits"),
                        )
              self.WriteLine(" ".join(command))
              self.WriteLine()
           self.dbdata.records[n].last = self.SkyMask

    def SkyModel(self,n,mask=None):
        if self.dbdata.records[n].last != self.SkyModel:
           image = self.dbdata.records[n].newname
           oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
           F = FITS(oimage,readonly=1)
           if self.rn: rn = self.rn
           else: rn = F["enoise"]
           g = F["egain"]
           sbinning = int(F["binning"].split("x")[0])
           binning = int(F["binning"].split("x")[1])
           F.close()
           skyint = 4096/(binning*2.0)
           kx, ky = 3, 0
           skymod = image.replace(".fits","m.fits")
#           if self.noslitfn: b = 1.0
#           else: b = 0
           b = 0
           pct = 40
           check = 5
           skycut = 4
           lowcut = 4
           spass = -5
#           if self.dbdata.records[self.spectra[0]].hilo:
           pct = 20
           skycut = 5
           lowcut = 5
           self.dkx = 1.00
           command = "skyrect",image
           if self.forked: command += "-forked",
           command += "-xdist","-ydist","-rn %f" % (rn), "-g ${GAIN}"
           command += "-skyint %d" % (skyint), "-kx %d" % (kx), "-ky %d" % (ky)
#           if not self.noobjects and not self.noskyaps:
#               command += "-skysuf sky",
#           if self.noobjects and not self.noskyaps:
           if not self.noskyaps:
               command += "-mask "+ self.dbdata.records[n].newname.replace(".fits","q.fits"),
           command += "-b %d" % (b), "-dkx %f" % (self.dkx),
           command += "-pct %f" % (pct), "-check %d" % (check),
           if mask: command += "-mask %s" % (mask),
           command += "-lowcut %f" % (lowcut), "-skycut %f" % (skycut), "-spass %d" % (spass)
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].last = self.SkyModel
           self.dbdata.records[n].newname = skymod

    def ScatModel(self,n):
        if self.dbdata.records[n].last != self.ScatModel:
           image = self.dbdata.records[n].newname
           scatmod = image.replace(".fits","m.fits")
           command = "dpScat",image
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].last = self.ScatModel
           self.dbdata.records[n].newname = scatmod

    def MakeSkyMask(self,n):
        image = self.dbdata.records[n].newname
        uimage = image.replace(".fits","r.fits")
        timage = lambda num: "/tmp/crap%d.fits" % (num)
        simage = image.replace(".fits","r_s.fits")
        uimage2 = image.replace(".fits","r_sd.fits")
        mimage = image.replace(".fits","r_sm.fits")
        command = "unrect -ydist -xdist -extend %s" % (image)
        self.WriteLine(command)
        command = "efits %s \"VTKMedianSmooth(i1,32,1,numret=1)\" %s" % (uimage,timage(1))
        self.WriteLine(command)
        command = "efits %s %s \"1.49*VTKMedianSmooth(abs(i1-i2),32,1,numret=1)\" %s" \
                   % (uimage,timage(1),timage(2))
        self.WriteLine(command)
        command = "efits %s %s \"divz(i1,i2)\" %s" % (uimage,timage(2),timage(3))
        self.WriteLine(command)
        command = "efits %s \"9999.*greater(VTKMeanSmooth(VTKConstantPad(VTKIslandRemoval(1.*greater(i1,1.0)*between(xy[1],4,2048-4),1.0,0.0,25,numret=1),(-1500,2048+1500,0,1499),0.0,numret=1),16,1,numret=1)*16,.125)[::,1500:2048+1500]\" %s" % (timage(3),simage)
        self.WriteLine(command)
        command = "rerect -ydist -xdist -extend -notdata -notcubic %s" % (simage)
        self.WriteLine(command)
        command = "efits %s \"1.*not_equal(i1,9999)\" %s" % (uimage2,mimage)
        self.WriteLine(command)
        self.dbdata.records[n].newname = image.replace("o.fits",".fits")
        self.dbdata.records[n].last = self.MakeSkyMask

    def SkyExposure(self,n,m):
        if self.dbdata.records[n].last != self.SkyExposure:
           image = self.dbdata.records[n].newname
           skymod = self.dbdata.records[m].newname
           command = "addkeys",image,"skyim",skymod
           command = join(command)
           self.WriteLine(command)
           image = image.replace(".fits","m.fits")
           command = "ln -fs",skymod,image
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].last = self.SkyExposure
           self.dbdata.records[n].newname = image

    def NewMask(self,n,jfit=0):
        if self.dbdata.records[n].last != self.NewMask:
           image = self.dbdata.records[n].newname
           oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
           F = FITS(oimage,readonly=1)
           if self.rn: rn = self.rn
           else: rn = F["enoise"]
           g = F["egain"]
           sbinning = int(F["binning"].split("x")[0])
           binning = int(F["binning"].split("x")[1])
           F.close()
           th = 10.0
           usemod = image.replace(".fits","v.fits")

           command = "efits",image,"\'VTKMedian(i1,5,5,numret=1)\'", usemod
           command = join(command)
           self.WriteLine(command)

           empnoise = image.replace(".fits","n.fits")
           command = "efits",image,usemod,"\'1.49*VTKMedian(abs(i1-i2),5,5,numret=1)\'", empnoise
           command = join(command)
           self.WriteLine(command)

           oldmask = image.replace(".fits","k.fits")
           sigmask = image.replace(".fits","z.fits")
           newmask = image.replace(".fits","l.fits")
           usemask = image.replace(".fits","u.fits")

           command = "efits",image,usemod,empnoise,"\'divz(i1-i2,i3)\'", sigmask
           command = join(command)
           self.WriteLine(command)

           command = "efits","-b 16",sigmask,oldmask,"\'i2*greater(VTKConvolve(less(abs(i1),%f).astype(float32),k=3,numret=1),0.75)\'" % (th), newmask
           command = join(command)
           self.WriteLine(command)

           command = "efits",image,newmask,"\'i1*i2\'",usemask
           command = join(command)
           self.WriteLine(command)
           self.WriteLine()

           if self.clean:
               command = "rm",oldmask,sigmask,usemask,empnoise,usemod
               command = join(command)
               self.WriteLine(command)
               self.WriteLine()

           self.dbdata.records[n].last = self.NewMask

    def SkySubtract(self,n,p=None):
        if self.dbdata.records[n].last != self.SkySubtract:
           skymod = self.dbdata.records[n].newname
           image = skymod.replace("m.fits",".fits")
           skysub = skymod.replace("m.fits","o.fits")
           if p == 2:
              command = "cp -a %s %s.orig" % (skysub,skysub)
              self.WriteLine(command)
           command = "efits",image,skymod,"\'(i1-i2)\'",skysub
           command = join(command)
           self.WriteLine(command)
           command = "addkeys",skysub,"skyim",skymod
           command = join(command)
           self.WriteLine(command)
           self.dbdata.records[n].newname = skysub
           self.dbdata.records[n].last = self.SkySubtract

    def PosObjects(self,spectra,ex=1.7625,oo=5):
        if self.dbdata.records[spectra[0]].last != self.PosObjects:
           self.apdb = "%s" % (self.target)
           if self.udate: self.apdb += "-%s" % (self.udate)
           command = "mkorderaps",
           if self.noobjects:
              command += "-dummy",
#              command += "-root %s-" % (self.dbdata.records[spectra[0]].newname.replace(".fits","")),
              command += "-root %s-" % (self.apdb),
           else:
              command += "-med",
              command += "-root %s-" % (self.apdb),
           if self.noskysub:
              command += "-ydist","-oo %d" % (oo),"-ex -nsig %f" % (ex),
           else:
              command += "-ydist","-bigsky","-oo %d" % (oo),"-ex -nsig %f" % (ex),
#           if self.sbinning == 1: edge = 3
#           else: edge = 1
           edge = 1
           command += "-edge %d" % (edge),
           for n in spectra:
              command += self.dbdata.records[n].newname,
              self.dbdata.records[n].last = self.PosObjects
           command = join(command)
           self.WriteLine(command)

    def AllPos(self,spectra):
        if self.dbdata.records[spectra[0]].last != self.AllPos:
           self.apdb = "%s.aps" % (self.apdb)
           command = "cat",
           for n in spectra:
              command += self.dbdata.records[n].newname.replace(".fits",".aps"),
              self.dbdata.records[n].last = self.AllPos
           command += ">", self.apdb
           command = join(command)
           self.WriteLine(command)

    def SumRect(self,kx=1,ky=1,xint=16,yint=64,fwhm=5,spass=3,mx=9,my=3,ncut=5.0,ccut=5.0):
        n = self.ispecs[0]
        oimage = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(oimage,readonly=1)
        if self.rn: rn = self.rn
        else: rn = F["enoise"]
        g = F["egain"]
        binning = int(F["binning"].split("x")[1])
        F.close()
        w1, w2 = 0, 4096/(binning*2.0)
        if self.twodim: command = "sumrect",
        else: command = "multiextrect",
        command += "-quiet","-xdist","-ydist","-wdist","-defsample",self.apdb
        if self.logb: command += "-logb",
        command += "-kx %d -ky %d" % (kx,ky),
        command += "-xint %d -yint %d" % (xint,yint),
        command += "-fwhm %f" % (fwhm), "-spass %d" % (spass),
        command += "-mx %d -my %d" % (mx,my), "-ncut %f" % (ncut), "-ccut %f" % (ccut),
        command += "-rn %f -g ${GAIN}" % (rn),
        command += "-extra %s" % (self.images[self.ixdists[0]]),
        command = join(command)
        self.WriteLine(command)

    def CRClean(self,n,th=40,mx=3,my=3):
        if self.dbdata.records[n].last != self.CRClean:
           image = self.dbdata.records[n].newname
           oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
           F = FITS(oimage,readonly=1)
           if self.rn: rn = self.rn
           else: rn = F["enoise"]
           g = F["egain"]
           sbinning = int(F["binning"].split("x")[0])
           binning = int(F["binning"].split("x")[1])
           F.close()
           command = "pyrclean -1",image,
#           command += "-bigsky",
           command += "-th %f" % (th),"-mx %d -my %d" % (mx,my)
           command += "-rn %f -g ${GAIN}" % (rn),
           command = join(command)
           self.WriteLine(command)
           cleaned = image.replace(".fits","c.fits")
           self.dbdata.records[n].last = self.CRClean
           self.dbdata.records[n].newname = cleaned

    def Unrect(self,spectra,fwhm=4,spass=1,mx=15,my=1,ncut=5.0,ccut=0.0,xsub=1,ysub=3,template=None):
        n = spectra[0]
        image = self.dbdata.records[n].newname
        extra = self.dbdata.records[n].uselamp
        extra = self.dbdata.records[extra].newname
        oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
        F = FITS(oimage,readonly=1)
        if self.rn: rn = self.rn
        else: rn = F["enoise"]
        g = F["egain"]
        binning = int(F["binning"].split("x")[1])
        F.close()
        if len(spectra) == 1 or self.noobjects:
           spass = 1
           ccut = ncut = 0.0
           mx = my = 1
        command = "multiunrect",
        if self.noobjects:
           apdb = self.dbdata.records[n].newname.replace(".fits",".aps")
           command += apdb,
           if template:
              timage = self.dbdata.records[template].newname.replace(".fits",".aps")
              command += "-template %s" % (timage),
           else:
              command += "-defsample",
        else:
           command += self.apdb,"-defsample",
        command += "-xdist","-ydist","-wdist",
        if self.logb: command += "-logb",
        command += "-fwhm %f" % (fwhm/binning), "-spass %d" % (spass),
        command += "-mx %d -my %d" % (mx,my), "-ncut %f" % (ncut), "-ccut %f" % (ccut),
        command += "-rn %f -g ${GAIN}" % (rn),
        command += "-xsub %d -ysub %d" % (xsub,ysub),
        command += "-extra %s" % (extra),
        if len(spectra) == 1: command += "-uniform",
        command = join(command)
        self.WriteLine(command)

    def CollateRect(self):
        self.root = "%s" % (self.target)
        if self.udate: self.root += "-%s" % (self.udate)
        command = "mikejoin","%s-???sum.fits" % (self.root)
        command += "-r","-o %s_sum.fits" % (self.root),
        command = join(command)
        self.WriteLine(command)
        command = "mikejoin","%s-???sumsky.fits" % (self.root)
        command += "-r","-o %s_sky.fits" % (self.root),
        command = join(command)
        self.WriteLine(command)
        command = "mikejoin","%s-???sumsig.fits" % (self.root)
        command += "-r","-o %s_sig.fits" % (self.root),
        command = join(command)
        self.WriteLine(command)
        command = "mikejoin","%s-???sumext.fits" % (self.root)
        command += "-r","-o %s_ext.fits" % (self.root),
        command = join(command)
        self.WriteLine(command)
        self.lasts = "%s_sum.fits" % (self.root),
        self.lasts += "%s_sky.fits" % (self.root),
        self.lasts += "%s_sig.fits" % (self.root),
        self.lasts += "%s_ext.fits" % (self.root),

    def MultiExtract(self,nsig=2.0):
        command = "extractmultispec","%s-???sum.fits" % (self.target)
        command += "-extra lamp","-nsig %f" % (nsig),"-o %s_multispec.fits" % (self.target),
        command += "-d %s_multispec.ps/vcps" % (self.target),
#        if self.clean: command += "-clean",
        command = join(command)
        self.WriteLine(command)
        self.lasts += "%s_multispec.fits" % (self.target),

    def MakeMask(self):
        maskimage = "mask.fits"
        command = "efits",self.useflat,"\"1.0*greater(i1,0.25)\"",maskimage
        command = join(command)
        self.WriteLine(command)

    def GHLBExtract(self,N=None,M=None):
        self.root = "%s" % (self.target)
        if N == None: n = self.spectra[0]
        else: n = N
        image = self.dbdata.records[n].newname
        oimage = join([self.dbdata.records[n].dirname, self.dbdata.records[n].basename],"/")
        F = FITS(oimage,readonly=1)
        if self.rn: rn = self.rn
        else: rn = F["enoise"]
        g = F["egain"]
        binning = int(F["binning"].split("x")[1])
        F.close()
        maskimage = "mask.fits"
        if self.dbdata.records[n].last != self.GHLBExtract:
           command = "time GHLBExtract",
           if self.fake: command += "-p",
           if N == None and len(self.spectra) > 1: command += "-npass 3 -nfactor 5",
           else: command += "-npass 1 -nfactor 0",
           command += "-rn %f -g ${GAIN}" % (rn),
           command += "-kx 1 -ky 0 -yx 4 -ya 8 -sx 2 -sa 2 -hy 2 -hx 2 -ha 2",
#           if self.dbdata.records[self.spectra[0]].hilo:
#              command += "-kx 5 -ky 0 -yx 6 -ya 6 -sx 2 -sa 2 -hy 8 -hx 4 -ha 2",
#           else:
#              command += "-kx 5 -ky 0 -yx 6 -ya 5 -sx 2 -sa 1 -hy 5 -hx 2 -ha 1",
           mlamp = self.dbdata.records[self.spectra[0]].uselamp
           command += "-lamp", self.dbdata.records[mlamp].newname,
           command += "-flat", self.useblaze
           command += "-mask",maskimage
           if N == None:
             command += "-o %s_multi.fits" % (self.root),
             command += "-header",self.dbdata.records[self.spectra[0]].simple
             for n in self.spectra:
                self.dbdata.records[n].last = self.GHLBExtract
#                command += self.dbdata.records[n].newname.replace("o.fits","o.aps"),
                command += self.dbdata.records[n].newname.replace(".fits",".aps"),
             self.lasts += "%s_multi.fits" % (self.root),
           else:
             self.dbdata.records[n].last = self.GHLBExtract
#             command += self.dbdata.records[n].newname.replace("o.fits","o.aps"),
             command += self.dbdata.records[n].newname.replace(".fits",".aps"),
             if M != None and N != M:
                command += "-template",self.dbdata.records[M].newname
                command += "-header",self.dbdata.records[n].newname
             self.dbdata.records[n].newname = self.dbdata.records[n].newname.replace("o.fits","o_multi.fits")
             command += "-o %s" % ( self.dbdata.records[n].newname),
             self.lasts += self.dbdata.records[n].newname,
           command = join(command)
           self.WriteLine(command)

    def ComputeMidpointTime(self,typeret):
        if typeret == "jd":
           F = FITS(self.dbdata.records[self.spectra[0]].filename)
           ra = ephem.hrrad(ephem.scanSexagesimal(F["ra"]))
           dec = ephem.degrad(ephem.scanSexagesimal(F["dec"]))
           F.close()
           mjd = [float(self.dbdata.records[n].mjd) for n in self.spectra]
           etime = [float(self.dbdata.records[n].etime) for n in self.spectra]
           mjd0 = mjd[0]
           mjd1 = mjd[-1] + etime[-1]/(3600.0*24.0)
           mjd_mid = (mjd0 + mjd1)/2.
           gjd_mid = 2400000.5 + mjd_mid
           hcp = ephem.heliocorr(gjd_mid,ra,dec)
           hjd_mid = gjd_mid - hcp
           return mjd_mid,hjd_mid
        elif typeret == "ut":
           etimes = [float(self.dbdata.records[n].etime) for n in self.spectra]
           utdate = [self.dbdata.records[n].date for n in self.spectra]
           utstart = [self.dbdata.records[n].start for n in self.spectra]
           dtstr0 = " ".join([utdate[0],utstart[0]])
           dtstr1 = " ".join([utdate[-1],utstart[-1]])
           dtobj0 = datetime.strptime(dtstr0, '%Y-%m-%d %H:%M:%S')
           dtobj1 = datetime.strptime(dtstr1, '%Y-%m-%d %H:%M:%S')
           dtstart = dtobj0
           dtend = dtobj1 + timedelta(seconds=etimes[-1])
           dtdiff = dtend - dtstart
           dtmid = dtstart + timedelta(seconds=dtdiff.seconds/2.)
           datemid,utmid = str(dtmid).split()
           dateend,utend = str(dtend).split()
           teff = float(dtdiff.seconds)
           return datemid,utmid,dateend,utend,teff
        else:
           print "Invalid choice for typeret! Exiting."
           sys.exit()

    def AddHeaderInfo(self,verb=None):
        if self.twodim:
           fr = "%s_multispec.fits" % (self.target)
        else:
           fr = "%s_multi.fits" % (self.target)
        mjd_mid,hjd_mid = self.ComputeMidpointTime(typeret="jd")
        date_mid,ut_mid,date_end,ut_end,teff = self.ComputeMidpointTime(typeret="ut")
        command = "addkeys -f %s mjd-mid %f" % (fr,mjd_mid)
        self.WriteLine(command)
        command = "addkeys -f %s hjd-mid %f" % (fr,hjd_mid)
        self.WriteLine(command)
        command = "addkeys %s ut-mid %s" % (fr,ut_mid)
        self.WriteLine(command)
        command = "addkeys %s date-mid %s" % (fr,date_mid)
        self.WriteLine(command)
        command = "addkeys %s ut-end %s" % (fr,ut_end)
        self.WriteLine(command)
        command = "addkeys %s date-end %s" % (fr,date_end)
        self.WriteLine(command)
        command = "addkeys -f %s efftime %s" % (fr,teff)
        self.WriteLine(command)

        command2 = "ecutils",
        if verb: command += "-v",
        command2 += "-f %s" % (fr),
        command2 = join(command2)
        self.WriteLine(command2)

    def CopyProducts(self):
        command = "cp",
        command += self.lasts
        command += "../Final-Products",
        command = join(command)
        self.WriteLine(command)

    def Report(self):
        command = ["mikeReport.py"]
        d = [d for d in sys.path if os.path.isdir(d) and command[0] in os.listdir(d)]
        if len(d):
            command = d[:1] + command
            command = join(command,"/")
            self.WriteLine(command)
            report = "%s_reportShort.pdf" % (self.root)
            command = "cp",report,"../Final-Products"
            command = join(command)
            self.WriteLine(command)

#            if self.owner:
#                command = ["ploneftp.py"]
#                d = [d for d in sys.path if os.path.isdir(d) and command[0] in os.listdir(d)]
#                if len(d):
#                    command = d[:1] + command
#                    command = [join(command,"/"),"-owner",self.owner,"-runid",self.runid]
#                    command = join(command)
#                    self.WriteLine(command)


