#!/usr/bin/env python

from SimpleQueue import *
from TwistedClient import *
from os import remove
from Expector import *
from VTKHelperFunctions import *
from pgplot import *
from basics import *
from numpy.core.numerictypes import *
import numpy
from FITS import *
from Multipack import *
from sets import *

def Ldecomp(w,lg,xco,yco):
    x = xco.ravel()
    y = yco.ravel()
    w = w.ravel()
    x0,x1 = extrema(x)
    xm = 0.5*(x1+x0)
    xr = 0.5*(x1-x0)
    xa = (x-xm)/xr
    lb = legendre(xa,lg)
    pl = linear_least_squares(transpose(lb*w),w*y)[0]
    return xm,xr,pl

def Hdecomp(ysh,ho,avs,yu,aqs,aws,ays,hw,rp,hodata=None):
    ju = nonzero(yu)
    nyu,avs,ays,aqs,aws,ays = take([ yu,avs,ays,aqs,aws,ays],ju,1)
    uays = (ays + ysh) / hw
    eb = getexp(-0.5*power(uays,2))
    if hodata:
       pb = hodata
       ho = len(pb)-1
       hb = hermite(uays,ho,norm=1)
    else:
       ew = divz(1,aqs)
       ew = ew/mean(compress(ew,ew))
       hb = hermite(uays,ho,norm=1)
#       pb = dot(hb*eb*nyu*ew,avs*nyu*ew)
       pb = linear_least_squares(transpose(hb*eb*nyu*ew),avs*nyu*ew)[0]
#       pb = svdfit(transpose(hb*eb*nyu*ew),avs*nyu*ew)
    hb0 = hermite(0.0,ho,norm=1)
    prof0 = dot(pb,hb0)
    pb = pb / prof0
    if rp:
       hb2 = zeros((ho+1,yu.shape[0]),float64)
       eb2 = zeros(yu.shape,float64)
       [put(hb2[ho1],ju,hb[ho1]) for ho1 in range(ho+1)]
       put(eb2,ju,eb)
       return pb,hb2,eb2
    else: return pb

class input_frame:
   def __init__(self,input,adu,rn,saturate=0.0,extra=None):
       print "Getting info for",input
       self.name = input
       if extra == None: extra = []
       if type(extra) == StringType: extra = [extra]
       self.extra = extra
       self.saturate = saturate
       f = FITS(input)
       self.adu = adu
       self.rn = rn
       self.skyim = f["skyim"]
       self.exptime = f["exptime"]
       self.utdate = f["ut-date"]
       if "MIKE" in f["instrume"]:
          self.utstart = f["ut-start"]
       else:
          self.utstart = f["ut-time"]
          self.utend = f["ut-end"]
       self.nslits = f["nslits"]
       self.naxis1 = f["naxis1"]
       self.naxis2 = f["naxis2"]
       self.ca = array([f["csect%da" % ( s+1)] for s in range(self.nslits)])
       self.cb = array([f["csect%db" % ( s+1)] for s in range(self.nslits)])
       self.apertures = [f["ecord%d" % (s)] for s in range(self.nslits)]
       self.objids  = [f["objid%d" % (s+1)].strip() for s in range(self.nslits)]
       if self.apertures[0] == "N/A": self.apertures = arange(self.nslits)+1
       else: self.apertures = asarray(self.apertures,int32)
       self.xrect = f["x-rect"]
       self.ydist = f["y-dist"]
       if f["w0xfmn"] != "N/A": self.havewaves = True
       else: self.havewaves = False
       f.close()
       ysol = readsol(self.name,"y0xfm")
       if len(ysol) == 5:
           dum1,dum2,self.ysol,self.yxord,self.yyord = ysol
           self.yzord = None
       elif len(ysol) == 6:
           dum1,dum2,self.ysol,self.yxord,self.yyord,self.yzord = ysol
       self.xdxx = arange(0,self.naxis1,1,float32)
   def __repr__(self): return self.name
   def GetDists(self,slit):
       dum1,dum2,self.xsol,self.xdord,self.ydord = readsol(self.name,"x%dxfm" % (slit))
       if self.havewaves:
          dum1,dum2,self.wsol,self.word,self.wbtype = readsol(self.name,"w%dxfm" % (slit))
       else:
          self.wsol = [array([self.naxis1/2.,self.naxis1/2.]),0.0]
          self.word = 1
          self.wbtype =  0
       self.xdn = [self.cb[slit] - self.ca[slit] + 0, self.naxis1]
   def ComputeYDist(self,y,slit):
       if self.yzord == None:
          dy = getmap(self.yxord,self.yyord,self.ysol[0],y*ones(self.xdxx.shape),self.xdxx,[self.naxis1,self.naxis2])
       else:
          xo = arange(self.naxis1).astype(float64)
          mo = self.nslits/2.0
          so = self.nslits/2.0
          xset = (xo-self.naxis1/2.)/(self.naxis1/2.)
          oset = ((slit-mo)/so)*ones(xo.shape,float64)
          xb = legendre(xset,self.yyord)
          ob = legendre(oset,self.yzord)
          b = asarray([xb1*ob1 for xb1 in xb for ob1 in ob])
          dy = dot(self.ysol[0],b)*self.naxis2
       return dy
   def ComputeXDist(self,y):
       xdyy = y
#       print type(y),y,Set(y),self.xdn
       dx = getmap(self.xdord,self.ydord,self.xsol[0],self.xdxx,xdyy*ones(self.xdxx.shape),self.xdn)
       return dx
   def ComputeWDist(self,xu):
       if self.wbtype == 1:
         xx = (2*(xu+1.0)-(1+self.naxis1))/(self.naxis1-1)
         xche = chebyshev(xx,len(self.wsol[0])-1)
         wx = add.reduce(self.wsol[0]*transpose(xche),1)
       else:
         x0,x1 = extrema(self.xdxx)
         xc = (x0 + x1)/2.0
         xw = (x1 - x0)/2.0
         xx = (xu-xc)/xw
         xleg = legendre(xx,len(self.wsol[0])-1)
         wx = add.reduce(self.wsol[0]*transpose(xleg),1)
       return wx
   def WhichSlit(self,loc,objid=None):
       if objid in self.objids:
          return self.objids.index(objid)
       else:
          bet = 1.0*between(loc,self.ca,self.cb)
          if sum(bet): return nonzero(between(loc,self.ca,self.cb))[0]
          else: return -99
   def ReadKeyword(self,keyword):
       f = FITS(self.name)
       value = f[keyword]
       f.close()
       return value
   def ReadData(self,yl,yh):
       print "Reading",self.name, [0,yl],[self.naxis1-1,yh-1]
       f = FITS(self.name)
       d = f.readsec([0,yl],[self.naxis1-1,yh-1])
       f.close()
       f = FITS(self.ydist)
       ydist = f.readsec([0,yl],[self.naxis1-1,yh-1])
       f.close()
       f = FITS(self.xrect)
       xrect = f.readsec([0,yl],[self.naxis1-1,yh-1])
       f.close()
       if self.skyim == "N/A": s = d
       else:
          f = FITS(self.skyim)
          s = f.readsec([0,yl],[self.naxis1-1,yh-1])
          f.close()
       e = []
       for iex in self.extra:
         print "Reading",iex
         f = FITS(iex)
         e.append( f.readsec([0,yl],[self.naxis1-1,yh-1]))
         f.close()
       return d,s,e,ydist,xrect
   def GetPixels(self,loc,hwid,lg=0,lgdata=None,lgran=None,raw=0,compute=1,edge=0,objid=None):
       t0 = time.time()
       print
       slit = self.WhichSlit(loc,objid=objid)
       if slit < -90: return None
       print "Aperture=",self.apertures[slit]
       a,b = self.ca[slit]-1+edge,self.cb[slit]-1-edge
       self.GetDists(slit)
       print "Slit boundaries = %d %d" % (a,b)
       print "Location = %.2f" % (loc)
       ys = asarray([loc-hwid, loc+hwid])
       ys = clip(ys,a,b)
       if ys[0] == a: ys[1] = min([ys[0]+2*hwid,b])
       if ys[1] == b: ys[0] = max([ys[1]-2*hwid,a])
       print "Extraction boundaries = %.2f %.2f" % (ys[0],ys[1])
       dys = asarray([self.ComputeYDist(y,slit) for y in ys])
       tys = (ys[::,NewAxis] + 1*dys)
       eys = (ys[::,NewAxis] + 2*dys)
       yl,yh = extrema(tys.ravel())
       yel,yeh = extrema(eys.ravel())
       yl = max([0,int(yl)])
       yel = max([0,int(yel)])+1
       yh = min([self.naxis2-1,int(yh+1.5)])
       yeh = min([self.naxis2-1,int(yeh+1.5)])+1
       nval,sval,extras,dyco,dxco = self.ReadData(yel,yeh)
       if self.saturate > 0:
          sval = where(less(nval,self.saturate),sval,0.0)
          nval = where(less(nval,self.saturate),nval,0.0)
       yco,xco = indices(nval.shape,float32)
       yco = add(yco,yel,yco)
       ayd = self.ComputeYDist(a,slit) + a
       lyd = self.ComputeYDist(loc,slit)
       byd = self.ComputeYDist(b,slit) + b
#       dyco = asarray([self.ComputeYDist(y,slit).astype(float32) for y in yco],float32)
       dxco = asarray([self.ComputeXDist(y-(a-0)).astype(float32) for y in yco],float32)
       xro = reversebin(xco,dxco,dyco,data=0,cubic=0,verb=0,constant=-999)
       yro = reversebin(yco,dxco,dyco,data=0,cubic=0,verb=0,constant=-999)
       cuse = 1.0*greater(xro,-100) * greater(yro,-100)
       cuse[:1] = 0 * cuse[:1]
       cuse[-1:] = 0 * cuse[-1:]
       wro = asarray([self.ComputeWDist(x) for x in xro])
       use = greater(wro,0)
       yro = yro - (loc - 0.0)
       use = use*between(yro,a-1-loc,b-1-loc,inc=0)*cuse
       use = VTKErode(use,3,3,numret=1)
       mhw = max([5.0,hwid])
       aduse = use*between(yro,-mhw,mhw)
       naduse = add.reduce(aduse,0)
       nadth = median(compress(naduse,naduse))-2
       nadlen = greater_equal(naduse,nadth)
       yuse = between(yco,ayd-1,byd+1.5,inc=1)
       nslits = len(self.ca)
       all_yuse = yuse
       for oslit in range(len(self.ca)):
#         ayd = 0*self.ComputeYDist(self.ca[oslit],slit) + self.ca[oslit] - 1 - edge - loc
#         byd = 0*self.ComputeYDist(self.cb[oslit],slit) + self.cb[oslit] - 1 + edge - loc
         ayd = self.ca[oslit] - 1 - edge - loc
         byd = self.cb[oslit] - 1 + edge - loc
         use = between(yro,ayd-1,byd+1.5,inc=1).astype(float32)
         all_yuse = all_yuse + use
       yuse = equal(all_yuse,2.0).astype(float32)
       lval = VTKLaplace(nval,numret=1)
       mval = VTKMedian(nval,25,1,numret=1)
       xpixarea = VTKGradMag(xro,1)
       ypixarea = VTKGradMag(yro,1)
       pixarea = xpixarea * ypixarea
       nval = VTKDivide(nval,pixarea)
       sval = VTKDivide(sval,pixarea)
       mval = VTKDivide(mval,pixarea)
       xval = [VTKDivide(extra,pixarea) for extra in extras]
#       if objid: qdump("yuse_%s.fits" % (objid),yuse*aduse*nval)
#       qdump("yuse.fits",all_yuse)
#       qdump("yuse.fits",yuse*nval)
#       qdump("nval.fits",nval)
#       qdump("mval.fits",mval)
#       qdump("sval.fits",sval)
#       exit()

       if raw:
          yuse = yuse*aduse*not_equal(sval,0)
          retvals = [compress(yuse.ravel(),b.ravel()) for b in [yro,yco,xro,xco,wro,nval,sval,mval,lval]]
          retvals += [compress(yuse.ravel(),b.ravel()) for b in xval]
#          order =  argsort(retvals[2])
#          retvals = take(retvals,order,1)
#          qdump("crap.fits",retvals[-1])
          return retvals

       if sum(lgdata) and lgran > 0:
           wcent = mean(lgran)
           wspan = diff(lgran)[0]/2.0
           lbase = (wro.ravel()-wcent)/wspan
           lbasis = legendre(lbase,lg)
           print "Using AD correction:",lgdata
           tb = dot(lgdata,lbasis)
           yro = reshape(yro.ravel() - tb,xro.shape)
       elif lg:
           duse = nval * aduse * nadlen
           muse = mval * aduse * nadlen
           suse = sval * aduse * nadlen
           vuse = (abs(suse)/self.adu+pow(self.rn/self.adu,2))*not_equal(suse,0)
#           wuse = divz(duse**2*greater(muse,-getsqrt(vuse)),vuse+muse*greater(muse,0)/self.adu+power(duse-muse,2))
           wuse = divz(duse*greater(muse,-getsqrt(vuse)),sqrt(vuse+muse*greater(muse,0)/self.adu+power(duse-muse,2)))
           wad0,wad1,ad = Ldecomp(wuse,lg,wro,yro)
           lb = legendre((wro.ravel()-wad0)/wad1,lg)
           print "Found ad=",ad
           tb = dot(ad,lb)
           yro = reshape(yro.ravel() - tb,xro.shape)

#       qdump("nval.fits",use*nval)
#       qdump("mval.fits",use*mval)
#       qdump("sval.fits",use*sval)
#       exit()

       print "Keeping object pixels"
#       use = use * not_equal(sval,0.0)
       use = yuse*aduse*greater(sval,0)
       ux,uw,uy,uv,um,us = compress(use.ravel(),[xro.ravel(),wro.ravel(),yro.ravel(),nval.ravel(),mval.ravel(),sval.ravel()],1)
       ue = [compress(use.ravel(),b) for b in xval]
       print "Sorting %d pixels by wavelength." % (len(uw))
       order = argsort(uw)
       ux,uw,uy,uv,um,us = take([ux,uw,uy,uv,um,us],order,1)
       ue = [take(order,b) for b in ue]
       t1 = time.time()
       print "Elapsed time = %.3f" % (t1-t0)
       print "Wavelength coverage = %.3fA -- %.3fA" % tuple(extrema(uw))
       print "Sums of object and sky pixels = %.1fDN and %.1fDN" % (sum(uv),sum(us))
       uq0 = abs(uv+us)/self.adu + pow(self.rn/self.adu,2)
       uq = abs(um+us)/self.adu + pow(self.rn/self.adu,2)
       uq = uq * greater(us,0)
       uq0 = uq0 * greater(us,0)
       self.doco = self.ComputeXDist(loc-a-1)
#       self.wxdxx = self.xdxx
#       self.wxdxx = self.xdxx + self.doco
       self.wxdxx = zeros( self.xdxx.shape,float64)
       derivx = asarray([diff(xro1) for xro1 in xro])
#       nxdxx = zeros( self.xdxx.shape,float64)
       nxdxx = []
       jxdxx = []
       for j in range(self.xdxx.shape[0]-1):
#          pxro, pyro = compress(use[::,j]*between(yro[::,j],-2*hwid,2*hwid,inc=1),[xro[::,j],yro[::,j]],1)
          nderix = compress(use[::,j]*between(yro[::,j],-2*hwid,2*hwid,inc=1),derivx[::,j])
          if len(nderix):
              nxdxx.append( median(nderix))
              jxdxx.append(j)
#             yz = [pxro[argmin(abs(pyro))]]
##             ytck = splrep(pxro,pyro,s=0,k=1)
##             yz = sproot(ytck,mest=3)
#             if len(yz):
#               nxdxx[j] = len(pxro)
#               self.wxdxx[j] = yz[0]
#       nxdxx,self.wxdxx = compress(self.wxdxx,[nxdxx,self.wxdxx],1)
#       mn = median(compress(nxdxx,nxdxx))
#       nxdxx = VTKDilate(nxdxx,15,1,1)
#       n0,n1 = take(nonzero(greater(nxdxx,mn-4)),[0,-1])
#       self.wxdxx = 1*self.wxdxx[n0:n1]
       jxdxx,nxdxx = asarray([jxdxx,nxdxx])
       ntck = splrep(jxdxx,nxdxx,s=0,k=3)
       self.wxdxx = [0.0]
       while self.wxdxx[-1] < self.xdxx[-1]-1:
          self.wxdxx.append( self.wxdxx[-1] + splev(self.wxdxx[-1],ntck))
       self.wxdxx = asarray( self.wxdxx )
#       qdump("w.fits",self.wxdxx)
#       qdump("wd.fits",diff(self.wxdxx))
#       qdump("n.fits",nxdxx)
#       exit()
       woco = self.ComputeWDist( self.wxdxx)
       kwpix = self.ComputeWDist( self.xdxx)
       kxpix = self.xdxx
       self.kxdxx = ones(woco.shape)
       xtck = splrep(kxpix,kwpix,s=0,k=3)
       dwdx = splev(kxpix,xtck,1)
       kpix = 1.0*greater(dwdx,0.0)
       xg0,xg1 = extrema(nonzero(kpix))
       use = between(ux,xg0,xg1,inc=1)
       use = use*greater(uq0,0.0)
       uw,uy,uv,um,us,uq,uq0 = compress(use, [uw,uy,uv,um,us,uq,uq0],1)
       ue = [compress(use,b) for b in ue]
       if len(ue): ue = ue[0]
       else: ue = zeros(uw.shape,float32)
       return woco,uw,uy,uv,um,us,ue,uq,uq0

def RobustFlux(dn):
   sdn = sort(dn)
   ldn = len(sdn)
   udn = sdn[int(ldn/4):int(ldn-ldn/40)]
   fdn = sum(udn)
   return fdn

class ObjectPackage:
    def __init__(self,names,locs,hwids,Inputs,ho=5,lg=0,subap=100,cut=5,ds=0,w1=0,w2=0,dw=0,logb=0,kx=3,ky=5,spass=1,sigkey=None,scalekey=None,ghkey=None,hodata=None,lgdata=None,lgran=None,insigs=None,inscales=None):
       self.names,self.locs,self.hwids,self.Inputs,self.ho,self.lg,self.subap,self.cut,self.ds,self.w1,self.w2,self.dw,self.logb,self.kx,self.ky,self.spass,self.sigkey,self.scalekey,self.ghkey,self.hodata,self.lgdata,self.lgran,self.insigs,self.inscales = \
               names,locs,hwids,Inputs,ho,lg,subap,cut,ds,w1,w2,dw,logb,kx,ky,spass,sigkey,scalekey,ghkey,hodata,lgdata,lgran,insigs,inscales
    def GetObject(self,k):
       print "Starting",self.names[k]
       status = GetObject(self.names[k],self.locs[k],self.hwids[k],self.Inputs[k],ho=self.ho,lg=self.lg,subap=self.subap,cut=self.cut,w1=self.w1,w2=self.w2,dw=self.dw,ds=self.ds,logb=self.logb,kx=self.kx,ky=self.ky,spass=self.spass,\
        sigkey=self.sigkey,scalekey=self.scalekey,ghkey=self.ghkey,hodata=self.hodata,lgdata=self.lgdata,lgran=self.lgran,insigs=self.insigs,inscales=self.inscales)

def GetObject(name,loc,hwid,Inputs,ho=5,lg=0,subap=100,cut=5,ds=0,w1=0,w2=0,dw=0,logb=0,kx=3,ky=5,spass=1,sigkey=None,scalekey=None,ghkey=None,hodata=None,lgdata=None,lgran=None,insigs=None,inscales=None):
   status = 0
   Nin = len(Inputs)
#   if Nin == 1: cut = 0.0
   print "Working on %s." % (name)
   scales = []
   uis = []
   uws = []
   uys = []
   uvs = []
   ums = []
   uss = []
   ues = []
   uqs = []
   uq0s = []
   wocos = []
   wsigs = []
   hoinfo = []
   hws = []
   juse = 0
   for j in range(Nin):
      if lgdata: lgd = lgdata[j]
      else: lgd = None
      if lgran: lgr = lgran[j]
      else: lgr = None
      getpix = Inputs[j].GetPixels(loc[j],hwid[j]+2*(ky+1),lg,lgdata=lgd,lgran=lgr)
      if getpix:
        juse += 1
        woco,uw,uy,uv,um,us,ue,uq,uq0 = getpix
        wocos.append( woco)
        hws.append(hwid[j])
        if juse == 1:
           xpix = Inputs[j].xdxx
           adu = Inputs[j].adu
           rn = Inputs[j].rn
        uis.append( juse*ones(uw.shape,float32))
        uws.append( uw)
        uys.append( uy)
        uvs.append( uv)
        ums.append( um)
        uss.append( us)
        ues.append( ue)
#        if Nin == 1 and cut == 0:
#        if Nin == 1:
#           uq = mean(uq.ravel())*ones(uq.shape,float32)
#           uq0 = mean(uq0.ravel())*ones(uq0.shape,float32)
        uqs.append( uq)
        uq0s.append( uq0)
        if ghkey:
          ho = Inputs[j].ReadKeyword("n%s" % (ghkey)) - 1
          hoinfo += [array([Inputs[j].ReadKeyword("see%s%d" % (ghkey,npbs)) for npbs in range(ho+1)])]
        else:
          if hodata: hoinfo.append( hodata[j])
          else: hoinfo.append(None)
        if sigkey: wsigs.append( Inputs[j].ReadKeyword(sigkey))
        elif insigs: wsigs.append( insigs[j])
        else:
           wavs = divz(uvs[-1],getsqrt(uq0s[-1]))**2
           wsig = hwid[j]/2.0
           dwsig = 1e9
           iwsig = 0
           while dwsig > 1e-2 and iwsig < 10:
             iwsig += 1
             uwsig = between(uys[-1],-2*wsig-1,2*wsig+1)*wavs
             wsig2 = sqrt(divz(sum(uwsig*power(uys[-1],2)),sum(uwsig)))
             dwsig = abs(wsig2-wsig)/wsig
             wsig = wsig2
           print "wsig1 =", wsig
           uuu = less(abs(uys[-1]),2*wsig)*greater(uqs[-1],0)*greater(ums[-1],0.0)
           uuy,uuv,uuw = compress(uuu,[uys[-1],uvs[-1],uq0s[-1]],1)
           uuw = 1/sqrt(uuw)
           def fitgp(params):
              s = 10**params[0]
              a = 10**params[1]
              p = a*getexp( -0.5*power(uuy/s,2))
              return (uuv-p)*(uuw)
           gfit = leastsq(fitgp,[log10(wsig),log10(getpct(uuv,80))],factor=0.01,xtol=1e-4,ftol=1e-4,gtol=1e-4,epsfcn=1e-2)[0]
           amp = 10**gfit[1]
           wsig_f = 10**gfit[0]
           print "amp_f=",amp
           print "wsig_f =", wsig_f
#           wsigs.append(min([wsig,wsig_f]))
           wsigs.append(wsig_f)
        if scalekey: scales.append( Inputs[j].ReadKeyword(scalekey))
        else:
          if inscales:
             scales.append( inscales[j])
          else:
             scale = sum(uv*between(uys[-1],-1.0*wsigs[-1],1.0*wsigs[-1],inc=1)*greater(um,0.0))
             scale = scale / sqrt(2*pi*wsigs[-1]**2)
             scales.append( scale)
        print "Sig = %.2f" % (wsigs[-1])
        print "Scale = %.2f" % (scales[-1])
   Nin = juse
   if not Nin: return
   wsigs = asarray(wsigs)
   scales = asarray(scales)
   scales = scales/mean(scales)
   print
   print "Scales = ",
   for s in scales: print "%.3f" % (s),
   print
   hwid = array(hws)
   hw = mean(scales*hwid)
   print "Using hwid = %.2f" % (hw)
   print
   t0 = time.time()
   for j in range(Nin):
      uqs[j] = uqs[j] / scales[j]**2
      uq0s[j] = uq0s[j] / scales[j]**2
      uvs[j] = uvs[j] / scales[j]
      ums[j] = ums[j] / scales[j]
      uss[j] = uss[j] / scales[j]
   ais = concatenate(uis)
   aws = concatenate(uws)
   ays = concatenate(uys)
   avs = concatenate(uvs)
   ams = concatenate(ums)
   ass = concatenate(uss)
   aes = concatenate(ues)
   aqs = concatenate(uqs)
   aq0s = concatenate(uq0s)
   ags = take(wsigs,ais.astype(int32)-1)
   del uis,uws,uys,uvs,ums,uss,ues,uqs,uq0s
   ais,aws,ays,avs,ams,ass,aes,aqs,aq0s = take( [ais,aws,ays,avs,ams,ass,aes,aqs,aq0s], argsort(aws), 1)
   wavs = divz(ams,getsqrt(aqs))**2
#   wsig = sqrt(divz(sum(between(ays,-hw-1,hw+1)*wavs*power(ays,2)),sum(between(ays,-hw-1,hw+1)*wavs)))
#   print "Total Light-weighted 2nd moment = %.2f pixels." % (wsig)
   wsigs = asarray(wsigs)
   print "Individual Light-weighted 2nd moments ="
   print wsigs
   wsig = sum(scales*wsigs)/sum(scales)
   print "Total Light-weighted 2nd moment = %.2f pixels." % (wsig)
   print
   woco = wocos[argmax(scales)]
   xall =  Inputs[argmax(scales)].xdxx
   wall =  Inputs[argmax(scales)].ComputeWDist( xall)
#   xpix = Inputs[argmax(scales)].wxdxx
#   wpix =  Inputs[argmax(scales)].ComputeWDist(xpix)
   xpix = xall
   wpix =  wall
   xtck = splrep(xpix,wpix,s=0,k=3)
   wtck = splrep(wpix,xpix,s=0,k=3)
   hws = 3*wsig
   if ky: npr = 2
   else: npr = spass
   for pa in range(npr):
       if pa == npr-1 and ky:
         fitted = splev(aws,dtck[0])
         dv = avs
         dq = aq0s
         dv = divz(avs,fitted)
         dq = divz(aq0s,fitted)
         prof = zeros(dv.shape,float64)
         pbs = []
         print "Gauss-Hermite-Legendre decomposition"
         for j in range(Nin):
             yv = equal(ais,j+1)
             pb,hb,eb = Hdecomp(0.0,ho,dv,yv,getsqrt(dq),aws,ays,wsigs[j],1,hodata=hoinfo[j])
             print "Normalized Coefficients (%d) =" % (j+1)
             print pb
             pbs.append(pb)
             prof1 = dot(pb,hb*eb)
             prof = where(yv,prof1,prof)
             if len(pb) > 1 and pb[1] > pb[0]: print "Failed on",name; status = -1; return status
         pbs = asarray(pbs)
         use = greater(prof,1e-8)
#         use = use*between(ays,-hw-2*(ky+1),hw+2*(ky+1),inc=1)
         ui,ux,uy,uf,us,uw,ue,uw0,up = compress(use, [ais,aws,ays,avs,ass,aes,getsqrt(aqs),getsqrt(aq0s),prof],1)

       if pa < npr-1 or ky == 0:
         if pa:
            fitted = splev(aws,dtck[0])
            dv = divz(avs,fitted)
            dq = divz(aq0s,fitted)
         else:
            dv = avs
            dq = aq0s
         print "Gauss-Hermite-Legendre decomposition"
         yu = between(ays,-2*hw-1*(ky+1),2*hw+1*(ky+1),inc=1)
         prof = zeros(dv.shape,float64)
         pbs = []
         for j in range(Nin):
            yv = equal(ais,j+1)
            pb,hb,eb = Hdecomp(0.0,ho,dv,yu*yv,getsqrt(dq),aws,ays,wsigs[j],1,hodata=hoinfo[j])
            print "Normalized Coefficients (%d) =" % (j+1)
            print pb
            pbs.append(pb)
            prof1 = dot(pb,hb*eb)
            prof = where(yv,prof1,prof)
            if len(pb) > 1 and pb[1] > pb[0]: print "Failed on",name; status = -1; return status

         use = between(ays,-hw,hw,inc=1)*greater(prof,1e-8)
         use = use*between(aws,woco[0],woco[-1],inc=1)

#         if cut and not ky:
         if cut:
            if pa:
               model = prof*fitted
               resid = model - avs
               signi = divz(resid,getsqrt(aq0s))
#               swid = 10*Nin*int(hw+ky) + 1
               swid = 20*int(hw+ky) + 1
               rsig = 1.25*divz(VTKGauss(abs(signi)*prof,swid,1), VTKGauss(1.*prof,swid,1))
               signr = divz(signi,rsig)
               print "Cutting=",cut
               use = use*less(abs(signr),cut)
            else:
#               swid = 4*Nin*(hw+ky)
               swid = 8*(hw+ky)
               smo = divz(VTKGauss(prof*divz(avs,getsqrt(aq0s)),swid,1), \
                       VTKGauss(prof*divz(prof,getsqrt(aq0s)),swid,1))
               model = prof*smo
               resid = model - avs
               signi = divz(resid,getsqrt(aq0s))
               swid = 10*int(swid)+1
               rsig = 1.25*divz(VTKGauss(abs(signi)*prof,swid,1), VTKGauss(1.*prof,swid,1))
               signr = divz(signi,rsig)
               print "Cutting=",cut
               use = use*less(abs(signr),cut)

         ui,ux,uy,uf,us,ue,uw,uw0,up = compress(use,[ais,aws,ays,avs,ass,aes,getsqrt(aqs),getsqrt(aq0s),prof],1)
         ux,ui,uy,uf,us,ue,uw,uw0,up = adjspac([ux,ui,uy,uf,us,ue,uw,uw0,up],0.001*median(diff(woco)))
         xb,xe = extrema(ux)
         tx = compress(between(woco,xb,xe),woco)
         if pa: uuw = uw0
         else: uuw = uw0
#         dtx = diff(tx)
#         dtx = concatenate([dtx,[dtx[-1]]])
         itx = searchsorted(ux,tx)
         ditx = diff(itx)
         utx = greater(ditx,1.0*(hw))
         tx = compress(utx,tx[1:])
         kx1 = [1,kx][ky==0]
         print "1D B-spline of order %d to %d pixels and %d knots." % (kx1, len(ux), len(tx))
         print "Range of knot spaces=",extrema(diff(tx))
#         print extrema(tx),extrema(ux)
#         tx=concatenate([(kx1)*[tx[0]],tx])
#         tx=concatenate([tx,(kx1)*[tx[-1]]])
#         while len(tx) < 2*(kx1+1):
#             tx = concatenate([[tx[0]],tx,[tx[-1]]])
         dtck = splrep(ux,uf/up,power(up,1.0)/uuw,s=None,task=-1,t=tx,xb=xb,xe=xe,k=kx1,full_output=1)
         msg = dtck[3]
         print msg
         ff = splev(ux,dtck[0])
         resid = ff - uf/up
         rsc = divz(resid,uw/up)
         chi2 = sum(power(rsc,2))
         rchi2 = chi2 / (len(uf)-len(dtck[0][1]))
         print "Chi2 = %.1f      Reduced Chi2 = %.4f" % (chi2,rchi2)
         t1 = time.time()
         print "Elapsed time = %.3f" % (t1-t0)
         if pa == npr-1:
            stck = splrep(ux,us,1/uuw,s=None,task=-1,t=tx,xb=xb,xe=xe,k=kx1,full_output=1)
            if sum(ue):
              etck = splrep(ux,ue,1/uuw,s=None,task=-1,t=tx,xb=xb,xe=xe,k=kx1,full_output=1)
         print
   if ds: w1,w2 = extrema(wall)
   if logb: w1,w2 = log10([w1,w2])
   if dw:
      if logb: nw = int((10**w2-10**w1)/dw+0.5)
      else: nw = int((w2-w1)/dw+0.5)
      dw = (w2-w1)/nw
   else:
      nw = len(xall)
      dw = (w2-w1)/nw
   g0,g1 = extrema(wall)
   wout = arange(nw)*dw + w1
   if logb: g0,g1 = log10([g0,g1])
   gout = between(wout,g0,g1)
   wout = clip(wout,g0,g1)
   if logb: rwout = power(10,wout)
   else: rwout = wout

   xout = splev(rwout,wtck)
   dxout = (xout[2:] - xout[:-2])/2.0
   dxout = concatenate([ [dxout[0]],dxout,[dxout[-1]]])

   dwout = (rwout[2:] - rwout[:-2])/2.0
   dwout = concatenate([ [dwout[0]],dwout,[dwout[-1]]])
   dxout = dxout*between(rwout,woco[0],woco[-1],inc=1)

   if ky == 0:
        pb = add.reduce(pbs*scales[::,NewAxis],0)/add.reduce(scales)
        dyho = 1.0/subap
        yho = arange(-hw,hw+dyho/2.,dyho)
        eb = exp(-0.5*power(yho/wsig,2))
        hb = hermite(yho/wsig,ho,norm=1)
        oprof = dot(pb,hb*eb)
        p2d = splev(rwout,dtck[0])[NewAxis,::]*oprof[::,NewAxis]
        s2d = splev(rwout,stck[0])[NewAxis,::]*ones(oprof[::,NewAxis].shape)
        if sum(ue): e2d = splev(rwout,etck[0])[NewAxis,::]*ones(oprof[::,NewAxis].shape)
        else: eout = False
        p2d = Nin*gout*dxout*p2d
        s2d = Nin*gout*dxout*s2d
        if sum(ue): e2d = Nin*gout*dxout*e2d
        poprof = power(oprof[::,NewAxis],0.0)
#        qdump("p2d.fits",p2d)
        pout = (2*hw)*add.reduce(p2d*poprof,0)/add.reduce(poprof,0)
        sout = (2*hw)*add.reduce(s2d*poprof,0)/add.reduce(poprof,0)
        if sum(ue): eout = (2*hw)*add.reduce(e2d*poprof,0)/add.reduce(poprof,0)
   else:
        pb = add.reduce(pbs*scales[::,NewAxis],0)/add.reduce(scales)
        dyho = 1.0/subap
        yho = arange(-hw,hw+dyho/2.,dyho)
        eb = exp(-0.5*power(yho/wsig,2))
        hb = hermite(yho/wsig,ho,norm=1)
        oprof = dot(pb,hb*eb)
        oprof2 = mean(oprof**2)
        oprof = oprof[::,NewAxis]
        p2d = zeros((yho.shape[0],rwout.shape[0]),float64)
        s2d = zeros((yho.shape[0],rwout.shape[0]),float64)
        n2d = zeros((yho.shape[0],rwout.shape[0]),float64)
        e2d = zeros((yho.shape[0],rwout.shape[0]),float64)
        use = between(aws,woco[0],woco[-1],inc=1)*greater(prof,1e-8)
        use = use * between(ays,-hw-2*(ky+1),hw+2*(ky+1))
        print extrema(aws),woco[0],woco[-1]
        ui0,ux0,uy0,uf0,us0,ue0,uw0,up0 = compress(use,[ais,aws,ays,avs,ass,aes,getsqrt(aq0s),prof],1)
        ui0,ux0,uy0,uf0,us0,ue0,uw0,up0 = \
                compress(not_equal(up0,0), [ui0,ux0,uy0,uf0,us0,ue0,uw0,up0], 1)

        if cut: npa = spass
        else: npa = 1
        woff = min([max([wsig,1.0]),hw])

        bs = nw/max([1,int(ui0.shape[0]/3e5)])
        nsec = int((nw-1)/bs)+1
        for isec in range(nsec):
           iws0 = isec*bs
           iws1 = min([(isec+1)*bs,len(rwout)])
           ws0,ws1 = extrema(wout[iws0:iws1])
#           if logb: ws0,ws1 = log10([ws0,ws1])
           print "Working #%d section from %d:%d." % (isec+1,iws0,iws1)
           if logb:
             ws0 = max([ws0 - 3*(kx+1)*dw,min(log10(ux0))])
             ws1 = min([ws1 + 3*(kx+1)*dw,max(log10(ux0))])
           else:
             ws0 = max([ws0 - 3*(kx+1)*dw,min(ux0)])
             ws1 = min([ws1 + 3*(kx+1)*dw,max(ux0)])
           if logb: ws0,ws1 = power(10,[ws0,ws1])
           print "And including wavelengths from %.3f to %.3f." % (ws0,ws1)
           use = between(ux0,ws0,ws1,inc=1)

           if sum(use):

              ui,ux,uy,uf,us,ue,uw,up = compress(use, [ui0,ux0,uy0,uf0,us0,ue0,uw0,up0], 1)
              xb,xe=extrema(ux)
              yb,ye=extrema(uy)

              for pa in range(npa):
#                  ukx,uky=kx,ky
                  if pa==npa-1: ukx,uky=kx,ky
                  else: ukx,uky=max([kx-2,1]),ky

                  nw = cut
                  if not cut or pa > 0:
                     uuw = uw
                  else:
                     print "Using modified weighting with the modification factor = ",cut
#                     umr = VTKMedian(uf/up,Nin*int(6*(ye-yb+1))+1,1,numret=1)
#                     ums = 1.49*VTKMedian(abs(uf/up-umr)/uw,Nin*int(6*(ye-yb+1))+1,1,numret=1)
                     umr = VTKMedian(uf/up,2*int(6*(ye-yb+1))+1,1,numret=1)
                     ums = 1.49*VTKMedian(abs(uf/up-umr)/uw,2*int(6*(ye-yb+1))+1,1,numret=1)
                     umn = divz(uf/up - umr,uw*ums)
                     uuw = uw*sqrt(cut**2+(umn)**2)

                  if not pa or 1:

#                     iux = arange(len(ux))
#                     ktck = splrep(iux,ux)
#                     tx = splev(arange(2*ky,len(ux)-2*ky,2*Nin*(hw+2*ky+2),float64),ktck)
#                     dtx = diff(tx)
#                     dtx = concatenate([dtx,dtx[-1:]])
#                     print "Extrema of knot spacings:",extrema(dtx)
                     tx = compress(between(woco,xb,xe,inc=1),woco)
                     itx = searchsorted(ux,tx)
                     ditx = diff(itx)
#                     utx = greater_equal(ditx,2*(hw+ky+1))
                     utx = greater_equal(ditx,(yb-ye))
                     tx = compress(utx,tx[1:])
                     dtx = diff(tx)
                     dtx = concatenate([dtx,dtx[-1:]])
                     print "Extrema of knot spacings:",extrema(dtx)

#                     tx=concatenate([(ukx+1)*[xb],tx])
#                     tx=concatenate([tx,(ukx+1)*[xe]])
#                     tx=concatenate([(ukx)*[tx[0]],tx])
#                     tx=concatenate([tx,(ukx)*[tx[-1]]])
                     ty=concatenate([(uky+1)*[-hw],(uky+1)*[hw]])

                  wup = power(up,1.0)
                  print "2D k=(%d,%d) B-spline to %d pixels." % (ukx,uky,len(ux))
                  dtck = bisplrep(ux,uy,uf/up,wup/uuw,s=None,task=-1,tx=tx,xb=xb,xe=xe,kx=ukx,ty=ty,yb=yb,ye=ye,ky=uky,nxest=len(tx),nyest=len(ty),full_output=1)
                  msg = dtck[3]
                  print msg
                  print "Getting the fit."
                  fitted0 = up0*array([bisplev(xy[0],xy[1],dtck[0]) for xy in transpose([ux0,uy0])])
                  fitted = compress(use,fitted0)
                  resid = fitted - uf
                  rescale = 1.49*median(divz(abs(resid),uw))
                  chi2 = sum(power(divz(resid,uw),2))
                  rchi2 = chi2 / (len(uf)-len(dtck[0][1]))
                  print "N = %7d    Chi2 = %.1f      Reduced Chi2 = %.4f Rescale = %.3f" % (resid.shape[0],chi2,rchi2,rescale)
                  if pa == npa-1:
                    print "2D B-spline to sky pixels."
                    stck = bisplrep(ux,uy,us,1/uuw,s=None,task=-1,tx=tx,xb=xb,xe=xe,kx=ukx,ty=ty,yb=yb,ye=ye,ky=uky,nxest=2*len(tx),nyest=2*len(ty),full_output=1,quite=1)
                    qtck = bisplrep(ux,uy,uw**2/up,wup/uuw,s=None,task=-1,tx=tx,xb=xb,xe=xe,kx=ukx,ty=ty,yb=yb,ye=ye,ky=uky,nxest=2*len(tx),nyest=2*len(ty),full_output=1,quite=1)
                    if sum(ue):
                       print "2D B-spline to extra pixels."
                       etck = bisplrep(ux,uy,ue,1/uuw,s=None,task=-1,tx=tx,xb=xb,xe=xe,kx=ukx,ty=ty,yb=yb,ye=ye,ky=uky,nxest=2*len(tx),nyest=2*len(ty),full_output=1,quite=1)
                  if pa < npa-1:
                     print "Getting the fit for the clipping."
                     resid = fitted0 - uf0
                     msc = divz(resid/up0,uw0/power(up0,1.0))
#                     qdump("msc.fits",msc)
                     print "Cutting at %.2f" % (nw)
                     if pa < npa-1:
                        use = 1.0*less(abs(msc),nw)
                        use = use * between(ux0,ws0,ws1,inc=1) * greater(up0,1e-8)
                        ui,ux,uy,uf,us,ue,uw,up = compress(use, [ui0,ux0,uy0,uf0,us0,ue0,uw0,up0], 1)
              print "Pasting into %d:%d" % (iws0,iws1)
              uwout = rwout[iws0:iws1]
              duwout = dwout[iws0:iws1]
#              if logb: uwout = power(10,uwout)
              tempor = transpose(bisplev(uwout,yho,dtck[0]))
              if len(tempor.shape) == 1: tempor = tempor[::,NewAxis]
              p2d[::,iws0:iws1] = tempor*oprof
              tempor = transpose(bisplev(uwout,yho,stck[0]))
              if len(tempor.shape) == 1: tempor = tempor[::,NewAxis]
              s2d[::,iws0:iws1] = tempor
              tempor = transpose(bisplev(uwout,yho,qtck[0]))
              if len(tempor.shape) == 1: tempor = tempor[::,NewAxis]
              n2d[::,iws0:iws1] = tempor*oprof
              if sum(ue):
                tempor = transpose(bisplev(uwout,yho,etck[0]))
                if len(tempor.shape) == 1: tempor = tempor[::,NewAxis]
                e2d[::,iws0:iws1] = tempor
                eout = True
              else: eout = False
              print
        p2d = Nin*gout*dxout*p2d
        s2d = Nin*gout*dxout*s2d
        n2d = Nin*gout*dxout*n2d
        if sum(ue): e2d = Nin*gout*dxout*e2d
        else: eout = False
        poprof = power(oprof,0)
        pout = (2*hw)*add.reduce(p2d*poprof,0)/add.reduce(poprof,0)
        sout = (2*hw)*add.reduce(s2d*poprof,0)/add.reduce(poprof,0)
        nout = (2*hw)*add.reduce(n2d*poprof,0)/add.reduce(poprof,0)
        if eout:
           eout = (2*hw)*add.reduce(e2d*oprof,0)/add.reduce(oprof,0)
   t1 = time.time()
#   qdump("p2d.fits",p2d)
#   qdump("s2d.fits",s2d)
   print "Elapsed time = %.3f" % (t1-t0)
   nout = getsqrt(abs(pout+sout)/(adu) + 2*hw*Nin*pow(rn/adu,2)*dxout)
   s2n = divz(pout,nout)
   sn11 = VTKGauss(divz(pout,nout),sigmax=2,sigmay=0,numret=1)
   sn12 = VTKGauss(divz(1.0,nout),sigmax=2,sigmay=0,numret=1)
   sn1 = divz(sn11,sn12)
   qout = divz(sn1,nout)
   WriteObject(name,Inputs[0].name,Nin,loc[0],hw,w1,dw,logb,adu,rn,\
        sout,pout,nout,qout,eout)
   t1 = time.time()
   print "Elapsed time = %.3f" % (t1-t0)
   print
   return status

def WriteObject(name,template,Nin,loc,hw,w1,dw,logb,adu,rn,sout,pout,nout,qout,eout):
   ename = name+"_1D.fits"
   f = FITS(ename,create=1)
   g = FITS(template)
   print "Creating",f.name()
   if eout: naxis2 = 5
   else: naxis2 = 4
   f.copyheader(g)
   f.modkey("naxis2",naxis2)
   f.resize([pout.shape[0],naxis2])
   f.close()
   nslits = int(g["nslits"])
   g.close()
   rmsol(ename,keyroot="e0xfm")
   rmsol(ename,keyroot="y0xfm")
   [rmsol(ename,keyroot="w%dxfm" % (oslit)) for oslit in range(nslits)]
   [rmsol(ename,keyroot="x%dxfm" % (oslit)) for oslit in range(nslits)]
#   f = FITS(ename,readonly=0)
#   [f.delkey("csect%da" % (oslit+1)) for oslit in range(nslits)]
#   [f.delkey("csect%db" % (oslit+1)) for oslit in range(nslits)]
#   [f.delkey("ecord%d" % (oslit)) for oslit in range(nslits)]
   f = FITS(ename,readonly=0)
#   f.delkey("nslits")
   f.modkey("xdcor",1)
   f.modkey("ydcor",1)
   f.modkey("nextract",2*hw)
   f.modkey("ncombine",Nin)
   f.modkey("crval1",w1)
   f.modkey("cdelt1",dw)
   f.modkey("crpix1",1)
   f.modkey("dc-flag",logb)
   f.modkey("eff-adu",adu)
   f.modkey("eff-rn",rn*sqrt(Nin))
   f.modkey("specd1","sky spectrum")
   f.modkey("specd2","object spectrum")
   f.modkey("specd3","noise spectrum")
   f.modkey("specd4","smoothed s/n ratio spectrum")
   if eout: f.modkey("specd5","extra spectrum")
   f.modkey("apsize",2*hw)
   f.modkey("aplow",loc-hw)
   f.modkey("aphigh",loc+hw)
   f.writerow(sout,0)
   f.writerow(pout,1)
   f.writerow(nout,2)
   f.writerow(qout,3)
   if eout: f.writerow(eout,4)
   f.close()

def WriteMultispec(D,Aps,CRVAL1,CDELT1,DCFLAG,APLOW,APHIGH,output,template,bandids):
    onaxes = list(D.shape)
    onaxes.reverse()
    Naps = D.shape[1]

    wat2 = ["wtype=multispec"]
    i = 0
    e = 0
    for n in range(Naps):
         ap = Aps[n]
         beam = e+1
         w1 = CRVAL1[n]
         dw = CDELT1[n]
         nw = onaxes[0]
         dtype = DCFLAG[0]
         z = 0.0
         aplow = APLOW[n]
         aphigh = APHIGH[n]
         wat2n = "spec%d = \"%d %d %d %r %r %d %f %f %f\"" % (i+1,ap,beam,dtype,w1,dw,nw,z,aplow,aphigh)
         wat2.append(wat2n)
         i += 1

    wat2 = join(wat2)
    ll = 68
    nwat2 = len(wat2)/ll + 1
    wat2 = [wat2[n*ll:(n+1)*ll] for n in range(nwat2)]

    O = FITS(output,readonly=0,create=1,template=template)
    if template: O.resize(onaxes)
    else: O.newhead("T",-32,len(onaxes),onaxes)
    O.delkey("crval1")
    O.delkey("crpix1")
    O.delkey("dc-flag")
    [O.modkey("ECORD%d" % (j),int(a)) for j,a in enumerate(Aps)]
    O.modkey("cdelt1",1.)
    O.modkey("cdelt2",1.)
    O.modkey("cdelt3",1.)
    O.modkey("cd1_1",1.)
    O.modkey("cd2_2",1.)
    O.modkey("cd3_3",1.)
    O.modkey("wcsdim",3)
    O.modkey("ltm1_1",1)
    O.modkey("ltm2_2",1)
    O.modkey("ltm3_3",1)
    O.modkey("ctype1","MULTISPE")
    O.modkey("ctype2","MULTISPE")
    O.modkey("ctype3","MULTISPE")
    [O.modkey("bandid%d" % (jb+1),bandid) for jb,bandid in enumerate(bandids)]
    O.modkey("wat0_001","system=multispec")
    O.modkey("wat1_001","wtype=multispec label=Wavelength units=angstroms")
    [O.modkey("wat2_%s" % (zfill(n+1,3)),wat2[n]) for n in range(len(wat2))]
    O.putdata(D)
    O.close()

def ReadMultispec(input,header=0,data=0):
    F = FITS(input)
    wat2 = "".join([F[k] for k in F.raw_keys() if "WAT2_" in k])
    if data: D = F.data()
    F.close()
    wat2 = wat2[wat2.find(" ")+1:].replace("spec",";spec")[1:]
    naps = wat2.count("spec")
    exec(wat2)
    apdata = transpose([eval("map(float,spec%d.split())" % (j)) for j in 1+arange(naps)])
    if header: return apdata
    elif data: return D

def WriteEquispec(D,objids,CRVAL1,CDELT1,DCFLAG,APLOW,APHIGH,output,template,bandids):
    if D.shape[1] == 1:
       D = MLab.squeeze(D)

    onaxes = list(D.shape)
    onaxes.reverse()
    Naps = D.shape[1]

    w1 = CRVAL1[0]
    dw = CDELT1[0]
    nw = onaxes[0]
    dtype = DCFLAG[0]

    if template:
       G = FITS(template)
       nslits = G["nslits"]
       G.close()

    O = FITS(output,readonly=0,create=1,template=template)
    if template:
       O.resize(onaxes)
       O.delkey("nslits")
       [O.delkey("csect%da" % (j+1)) for j in range(nslits)]
       [O.delkey("csect%db" % (j+1)) for j in range(nslits)]
    else:
        O.newhead("T",-32,len(onaxes),onaxes)
    O.modkey("crval1",w1)
    O.modkey("crval2",1.)
    if len(onaxes) == 3:
       O.modkey("crval3",1.)
    O.modkey("crpix1",1.)
    O.modkey("crpix2",1.)
    if len(onaxes) == 3:
       O.modkey("crpix3",1.)
    O.modkey("cdelt1",dw)
    O.modkey("cdelt2",1.)
    if len(onaxes) == 3:
       O.modkey("cdelt3",1.)
    O.modkey("cd1_1",dw)
    O.modkey("cd2_2",1.)
    if len(onaxes) == 3:
       O.modkey("cd3_3",1.)
    O.modkey("wcsdim",len(onaxes))
    O.modkey("dispaxis",1)
    O.modkey("dc-flag",dtype)
    O.modkey("ltm1_1",1)
    O.modkey("ltm2_2",1)
    if len(onaxes) == 3:
       O.modkey("ltm3_3",1)
    O.modkey("ctype1","LINEAR")
    O.modkey("ctype2","LINEAR")
    if len(onaxes) == 3:
       O.modkey("ctype3","LINEAR")
    [O.modkey("bandid%d" % (jb+1),bandid) for jb,bandid in enumerate(bandids)]
    O.modkey("wat0_001","system=equispec")
    O.modkey("wat1_001","wtype=linear label=Wavelength units=angstroms")
    O.modkey("wat2_001","wtype=linear")
    O.modkey("wat3_001","wtype=linear")
    [O.modkey("apnum%d" % (j+1),"%d %s %.2f %.2f" % (j+1,objids[j],APLOW[j],APHIGH[j])) for j in range(len(APLOW))]
    O.putdata(D)
    O.close()

